#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Fake client exceptions to use."""


class UnsupportedVersion(Exception):
    """Unsupported version of the client."""
    pass


class ClientException(Exception):
    """The base exception class for these fake exceptions."""
    _error_code = None
    _error_desc = None
    _error_ref = None

    _debug1 = None
    _debug2 = None

    def __init__(self, error=None):
        if error:
            if 'code' in error:
                self._error_code = error['code']
            if 'desc' in error:
                self._error_desc = error['desc']
            if 'ref' in error:
                self._error_ref = error['ref']

            if 'debug1' in error:
                self._debug1 = error['debug1']
            if 'debug2' in error:
                self._debug2 = error['debug2']

    def get_code(self):
        return self._error_code

    def get_description(self):
        return self._error_desc

    def get_ref(self):
        return self._error_ref

    def __str__(self):
        formatted_string = self.message
        if self.http_status:
            formatted_string += " (HTTP %s)" % self.http_status
        if self._error_code:
            formatted_string += " %s" % self._error_code
        if self._error_desc:
            formatted_string += " - %s" % self._error_desc
        if self._error_ref:
            formatted_string += " - %s" % self._error_ref

        if self._debug1:
            formatted_string += " (1: '%s')" % self._debug1

        if self._debug2:
            formatted_string += " (2: '%s')" % self._debug2

        return formatted_string


class HTTPConflict(ClientException):
    http_status = 409
    message = "Conflict"

    def __init__(self, error=None):
        if error:
            super(HTTPConflict, self).__init__(error)
            if 'message' in error:
                self._error_desc = error['message']

    def get_description(self):
        return self._error_desc


class HTTPNotFound(ClientException):
    http_status = 404
    message = "Not found"


class HTTPForbidden(ClientException):
    http_status = 403
    message = "Forbidden"


class HTTPBadRequest(ClientException):
    http_status = 400
    message = "Bad request"


class HTTPUnauthorized(ClientException):
    http_status = 401
    message = "Unauthorized"


class HTTPServerError(ClientException):
    http_status = 500
    message = "Error"

    def __init__(self, error=None):
        if error and 'message' in error:
            self._error_desc = error['message']

    def get_description(self):
        return self._error_desc



# Copyright (C) 2016-2022 Lightbits Labs Ltd.
# Copyright (C) 2020 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client as httpstatus
import json
import random
import time
from typing import Dict
from urllib.parse import urlparse


from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units
import requests
import urllib3

from cinder.common import constants
from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume import configuration as config
from cinder.volume import driver


LOG = logging.getLogger(__name__)
ENABLE_TRACE = True
LIGHTOS_DEFAULT_PROJECT_NAME = "default"

urllib3.disable_warnings()

lightos_opts = [
    cfg.ListOpt('lightos_api_address',
                default=None,
                item_type=cfg.types.IPAddress(),
                help='The IP addresses of the LightOS API servers separated'
                ' by commas.'),
    cfg.PortOpt('lightos_api_port',
                default='443',
                help='The TCP/IP port at which the LightOS API'
                     ' endpoints listen.'
                     ' Port 443 is used for HTTPS and other values'
                     ' are used for HTTP.'),
    cfg.StrOpt('lightos_jwt',
               default=None,
               help='JWT to be used for volume and snapshot operations with'
                    ' the LightOS cluster.'
                    ' Do not set this parameter if the cluster is installed'
                    ' with multi-tenancy disabled.'),
    cfg.IntOpt('lightos_default_num_replicas',
               min=1,
               max=3,
               default=3,
               help='The default number of replicas to create for each'
                    ' volume.'),
    cfg.BoolOpt('lightos_default_compression_enabled',
                default=False,
                help='Set to True to create  new volumes compressed assuming'
                     ' no other compression setting is specified via the'
                     ' volumes type.'),
    cfg.IntOpt('lightos_api_service_timeout',
               default=30,
               help='The default amount of time (in seconds) to wait for'
               ' an API endpoint response.')
]

CONF = cfg.CONF
CONF.register_opts(lightos_opts, group=config.SHARED_CONF_GROUP)
BLOCK_SIZE = 8
LIGHTOS = "LIGHTOS"
INTERM_SNAPSHOT_PREFIX = "for_clone_"


class LightOSConnection(object):
    def __init__(self, conf):
        self.conf = conf
        self.access_key = None
        self.apiservers = self._init_api_servers()
        self._cur_api_server_idx = random.randint(0, len(self.apiservers) - 1)
        self.targets = dict()
        self.lightos_cluster_uuid = None
        self.subsystemNQN = None
        self._stats = {'total_capacity_gb': 0, 'free_capacity_gb': 0}
        # a single API call must have been answered in this time if the API
        # service/network were up
        self.api_timeout = self.conf.lightos_api_service_timeout

    def _init_api_servers(self) -> Dict[int, Dict]:
        # And verify that port is in range
        apiservers: Dict[int, Dict] = {}
        hosts = self.conf.lightos_api_address
        port = str(self.conf.lightos_api_port)
        apiservers = [dict(api_address=addr, api_port=port) for addr in hosts]
        return apiservers

    def _generate_lightos_cmd(self, cmd, **kwargs):
        """Generate command to be sent to LightOS API service"""

        def _joined_params(params):
            param_str = []
            for k, v in params.items():
                param_str.append("%s=%s" % (k, v))
            return '&'.join(param_str)

        # Dictionary of applicable LightOS commands in the following format:
        # 'command': (method, API_URL, {optional parameters})
        # This is constructed on the fly to include the caller-supplied kwargs
        # Can be optimized by only constructing the specific
        # command the user provided in cmd

        # API V2 common commands
        lightos_commands = {
            # cluster operations,
            'get_cluster_info': ('GET',
                                 '/api/v2/clusterinfo', {}),

            'get_cluster': ('GET',
                            '/api/v2/cluster', {}),

            # node operations
            'get_node': ('GET',
                         '/api/v2/nodes/%s' % kwargs.get('UUID'), {}),

            'get_nodes': ('GET',
                          '/api/v2/nodes', {}),

            # volume operations
            'create_volume': ('POST',
                              '/api/v2/projects/%s/volumes' % kwargs.get(
                                  "project_name"),
                              {
                                  'name': kwargs.get('name'),
                                  'size': kwargs.get('size'),
                                  'replicaCount': kwargs.get('n_replicas'),
                                  'compression': kwargs.get('compression'),
                                  'acl': {
                                      'values': kwargs.get('acl'),
                                  },
                                  'sourceSnapshotUUID': kwargs.get(
                                      'src_snapshot_uuid'),
                                  'sourceSnapshotName': kwargs.get(
                                      'src_snapshot_name'),
                              }),

            'delete_volume': ('DELETE',
                              '/api/v2/projects/%s/volumes/%s' % (kwargs.get(
                                  "project_name"), kwargs.get("volume_uuid")),
                              {}),

            'update_volume': ('PUT',
                              '/api/v2/projects/%s/volumes/%s' % (kwargs.get(
                                  "project_name"), kwargs.get("volume_uuid")),
                              {
                                  'acl': {
                                      'values': kwargs.get('acl'),
                                  },
                              }),

            'extend_volume': ('PUT',
                              '/api/v2/projects/%s/volumes/%s' % (
                                  kwargs.get("project_name"),
                                  kwargs.get("volume_uuid")),
                              {
                                  'UUID': kwargs.get('volume_uuid'),
                                  'size': kwargs.get('size'),
                              }),

            # snapshots operations
            'create_snapshot': ('POST',
                                '/api/v2/projects/%s/snapshots' % kwargs.get(
                                    "project_name"),
                                {
                                    'name': kwargs.get('name'),
                                    'sourceVolumeUUID': kwargs.get(
                                        'src_volume_uuid'),
                                    'sourceVolumeName': kwargs.get(
                                        'src_volume_name'),
                                }),

            'delete_snapshot': ('DELETE',
                                '/api/v2/projects/%s/snapshots/%s' % (
                                    kwargs.get("project_name"),
                                    kwargs.get("snapshot_uuid")),
                                {}),

            # get operations
            'get_volume': ('GET',
                           '/api/v2/projects/%s/volumes/%s' % (
                               kwargs.get("project_name"),
                               kwargs.get("volume_uuid")),
                           {}),

            'get_volume_by_name': ('GET',
                                   '/api/v2/projects/%s/volumes/?name=%s' % (
                                       kwargs.get("project_name"),
                                       kwargs.get("volume_name")),
                                   {}),

            'list_volumes': ('GET',
                             '/api/v2/projects/%s/volumes' % kwargs.get(
                                 "project_name"),
                             {}),

            'get_snapshot': ('GET',
                             '/api/v2/projects/%s/snapshots/%s' % (
                                 kwargs.get("project_name"),
                                 kwargs.get("snapshot_uuid")),
                             {}),

            'get_snapshot_by_name': ('GET',
                                     '/api/v2/projects/%s/snapshots'
                                     '/?Name=%s' % (
                                         kwargs.get("project_name"),
                                         kwargs.get("snapshot_name")),
                                     {})
        }
        if cmd not in lightos_commands:
            raise exception.UnknownCmd(cmd=cmd)
        else:
            (method, url, params) = lightos_commands[cmd]

        if method == 'GET':
            body = params

        elif method == 'DELETE':
            LOG.debug("DELETE params: %s", params)
            # For DELETE commands add parameters to the URL
            url += '?' + _joined_params(params)
            body = ''

        elif method == 'PUT':
            # For PUT commands add parameters to the URL
            body = params

        elif method == 'POST':
            body = params

        else:
            msg = (_('Method %(method)s is not defined') %
                   {'method': method})
            LOG.error(msg)
            raise AssertionError(msg)

        return (method, url, body)

    def pretty_print_req(self, req, timeout):
        request = req.method + ' ' + req.url
        header = ', '.join('"{}: {}"'.format(k, v)
                           for k, v in req.headers.items())
        LOG.debug('Req: %s Headers: %s Body: %s Timeout: %s',
                  request,
                  header,
                  req.body,
                  timeout)

    def send_cmd(self, cmd, timeout, **kwargs):
        """Send command to any LightOS REST API server."""
        start_idx = self._cur_api_server_idx
        stop = time.time() + timeout
        while time.time() <= stop:
            server = self.apiservers[self._cur_api_server_idx]
            host = server['api_address']
            port = server['api_port']

            (success, status_code, data) = self.__send_cmd(
                cmd, host, port, self.api_timeout, **kwargs)
            if success:
                return (status_code, data)
            # go on to the next API server wrapping around as needed
            self._cur_api_server_idx = (
                self._cur_api_server_idx + 1) % len(self.apiservers)
            # if we only have a single API server, keep trying it
            # if we have more than one and we tried all of them, give up
            if (self._cur_api_server_idx ==
                    start_idx and len(self.apiservers) > 1):
                break

        raise exception.VolumeDriverException(
            message="Could not get a response from any API server")

    def _format_endpoint(self, ip, port):
        ip_requires_bracketing = ':' in ip or '%' in ip
        template = "[%s]:%s" if ip_requires_bracketing else "%s:%s"
        return template % (ip, port)

    def __send_cmd(self, cmd, host, port, timeout, **kwargs):
        """Send command to LightOS REST API server.

        Returns: (success = True/False, data)
        """
        ssl_verify = self.conf.driver_ssl_cert_verify
        (method, url, body) = self._generate_lightos_cmd(cmd, **kwargs)
        LOG.info(
            'Invoking %(cmd)s using %(method)s url: %(url)s \
            request.body: %(body)s ssl_verify: %(ssl_verify)s',
            {'cmd': cmd, 'method': method, 'url': url, 'body': body,
             'ssl_verify': ssl_verify})

        api_url = "https://%s%s" % (self._format_endpoint(host, port), url)

        try:
            with requests.Session() as session:
                req = requests.Request(
                    method, api_url, data=json.dumps(body) if body else None)
                req.headers.update({'Accept': 'application/json'})
                # -H 'Expect:'  will prevent us from getting
                # the 100 Continue response from curl
                req.headers.update({'Expect': ''})
                if method in ('POST', 'PUT'):
                    req.headers.update({'Content-Type': 'application/json'})
                if kwargs.get("etag"):
                    req.headers.update({'If-Match': kwargs['etag']})
                if self.conf.lightos_jwt:
                    req.headers.update(
                        {'Authorization':
                         'Bearer %s' % self.conf.lightos_jwt})
                prepped = req.prepare()
                self.pretty_print_req(prepped, timeout)
                response = session.send(
                    prepped, timeout=timeout, verify=ssl_verify)
        except Exception:
            LOG.exception("REST server not responding at '%s'", api_url)
            return (False, None, None)

        try:
            resp = response.json()
        except ValueError:
            resp = response.text
        data = resp

        LOG.debug(
            'Resp(%s): code %s data %s',
            api_url,
            response.status_code,
            data)
        return (True, response.status_code, data)


@interface.volumedriver
class LightOSVolumeDriver(driver.VolumeDriver):
    """OpenStack NVMe/TCP cinder drivers for Lightbits LightOS.

    .. code-block:: default

      Version history:
          2.3.12 - Initial upstream driver version.
    """

    VERSION = '2.3.12'
    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "LightbitsLabs_CI"

    def __init__(self, *args, **kwargs):
        super(LightOSVolumeDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(lightos_opts)
        # connector implements NVMe/TCP initiator functionality.
        if not self.configuration.__dict__.get("initiator_connector", None):
            self.configuration.initiator_connector = (
                "os_brick.initiator.connector.InitiatorConnector")
        if not self.configuration.__dict__.get("lightos_client", None):
            self.configuration.lightos_client = (
                "cinder.volume.drivers.lightos.LightOSConnection")

        initiator_connector = importutils.import_class(
            self.configuration.initiator_connector)
        self.connector = initiator_connector.factory(
            LIGHTOS,
            root_helper=utils.get_root_helper(),
            message_queue=None,
            device_scan_attempts=
            self.configuration.num_volume_device_scan_tries)

        lightos_client_ctor = importutils.import_class(
            self.configuration.lightos_client)
        self.cluster = lightos_client_ctor(self.configuration)

        self.logical_op_timeout = \
            self.configuration.lightos_api_service_timeout * 3 + 10

    @classmethod
    def get_driver_options(cls):
        additional_opts = cls._get_oslo_driver_opts(
            'driver_ssl_cert_verify', 'reserved_percentage',
            'volume_backend_name')
        return lightos_opts + additional_opts

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume.

        If volume_type extra specs includes 'replication: <is> True' the
        driver needs to create a volume replica (secondary)
        and setup replication between the newly created volume
        and the secondary volume.
        """
        project_name = self._get_lightos_project_name(volume)
        # Create an intermediate snapshot
        snapshot_name = self._interm_snapshotname(volume)
        src_volume_name = self._lightos_volname(src_vref)
        try:
            self._create_snapshot(project_name,
                                  snapshot_name, src_volume_name)
        except Exception as e:
            LOG.warning(
                "Failed to create intermediate snapshot \
                 %s from source volume %s.",
                snapshot_name,
                src_volume_name)
            raise e

        # Create a volume from the intermediate snapshot
        try:
            self._create_volume(volume,
                                src_snapshot_lightos_name=snapshot_name)
        except Exception as e:
            LOG.error("Failed to create volume %s from intermediate "
                      " snapshot %s. Trying to clean up.",
                      src_volume_name, snapshot_name)
            raise e

        # Delete the intermediate snapshot
        finally:
            try:
                self._delete_lightos_snapshot(project_name, snapshot_name)
            except Exception as e:
                LOG.warning("Failed to delete the intermediate snapshot %s for"
                            " volume %s. Trying to clean up.",
                            snapshot_name, src_volume_name)
                raise e

    def create_export(self, context, volume, vg=None):
        """Irrelevant for lightos volumes.

        Export created during attachment.
        """
        pass

    def ensure_export(self, context, volume):
        """Irrelevant for lightos volumes.

        Export created during attachment.
        """
        pass

    def remove_export(self, context, volume):
        """Irrelevant for lightos volumes.

        Export removed during detach.
        """
        pass

    def _get_lightos_volume(
            self,
            project_name,
            timeout,
            vol_uuid=None,
            vol_name=None):
        assert vol_uuid or vol_name, 'LightOS volume name or UUID \
        must be specified'
        if vol_uuid:
            return self.cluster.send_cmd(
                cmd='get_volume',
                project_name=project_name,
                timeout=timeout,
                volume_uuid=vol_uuid)

        return self.cluster.send_cmd(
            cmd='get_volume_by_name',
            project_name=project_name,
            timeout=timeout,
            volume_name=vol_name)

    def _lightos_volname(self, volume):
        volid = volume.name_id
        lightos_volname = CONF.volume_name_template % volid
        return lightos_volname

    def _get_lightos_project_name(self, volume):
        try:
            extra_specs = volume.volume_type.extra_specs
            project_name = extra_specs.get(
                'lightos:project_name',
                LIGHTOS_DEFAULT_PROJECT_NAME)
        except Exception:
            LOG.debug(
                "LIGHTOS volume %s has no lightos:project_name",
                volume)
            project_name = LIGHTOS_DEFAULT_PROJECT_NAME

        return project_name

    def _lightos_snapshotname(self, snapshot_id):
        return CONF.snapshot_name_template % snapshot_id

    def _interm_snapshotname(self, snapshot):
        id = snapshot['id']
        return '%s%s' % (INTERM_SNAPSHOT_PREFIX, id)

    def _get_lightos_snapshot(
            self,
            project_name,
            timeout,
            snapshot_uuid=None,
            snapshot_name=None):
        assert snapshot_uuid or snapshot_name, 'LightOS snapshot name or \
        UUID must be specified'
        if snapshot_uuid:
            return self.cluster.send_cmd(
                cmd='get_snapshot',
                project_name=project_name,
                timeout=timeout,
                snapshot_uuid=snapshot_uuid)

        return self.cluster.send_cmd(
            cmd='get_snapshot_by_name',
            project_name=project_name,
            timeout=timeout,
            snapshot_name=snapshot_name)

    def _wait_for_volume_available(
            self,
            project_name,
            timeout,
            vol_uuid=None,
            vol_name=None):
        """Wait until the volume is available."""
        assert vol_uuid or vol_name, 'LightOS volume UUID or name \
        must be supplied'
        # while creating lightos volume we can stop on any terminal status
        # possible states: Unknown, Creating, Available, Deleting, Deleted,
        # Failed, Updating, Migrating, Rollback
        states = ('Available', 'Deleting', 'Deleted', 'Failed', 'UNKNOWN',
                  'Migrating', 'Rollback')

        stop = time.time() + timeout
        while time.time() <= stop:
            (status_code,
             resp) = self._get_lightos_volume(project_name,
                                              timeout=self.logical_op_timeout,
                                              vol_uuid=vol_uuid,
                                              vol_name=vol_name)
            state = resp.get('state', 'UNKNOWN') if \
                status_code == httpstatus.OK and resp else 'UNKNOWN'
            if state in states and status_code != httpstatus.NOT_FOUND:
                break
            time.sleep(1)

        return state

    def _parse_extra_spec(self, extra_spec_value, default_value):
        extra_spec_value = str(extra_spec_value)
        extra_spec_value = extra_spec_value.casefold()
        if "true" in extra_spec_value:
            return "True"
        elif "false" in extra_spec_value:
            return "False"
        return default_value

    def _get_volume_specs(self, volume):
        default_compression = 'True' if self.configuration. \
            lightos_default_compression_enabled else 'False'
        num_replicas = str(self.configuration.lightos_default_num_replicas)

        if not volume.volume_type:
            return (default_compression, num_replicas,
                    LIGHTOS_DEFAULT_PROJECT_NAME)

        specs = getattr(volume.volume_type, 'extra_specs', {})
        type_compression = specs.get('compression', default_compression)
        compression = self._parse_extra_spec(type_compression,
                                             default_compression)
        num_replicas = str(specs.get('lightos:num_replicas', num_replicas))
        project_name = specs.get(
            'lightos:project_name',
            LIGHTOS_DEFAULT_PROJECT_NAME)
        return (compression, num_replicas, project_name)

    def _create_new_lightos_volume(self,
                                   os_volume,
                                   project_name,
                                   lightos_name,
                                   src_snapshot_lightos_name=None):
        """Create a new LightOS volume for this openstack volume."""
        (compression, num_replicas, _) = self._get_volume_specs(os_volume)
        return self.cluster.send_cmd(
            cmd='create_volume',
            project_name=project_name,
            timeout=self.logical_op_timeout,
            name=lightos_name,
            size=str(os_volume['size']) + ' gib',
            n_replicas=num_replicas,
            compression=compression,
            src_snapshot_name=src_snapshot_lightos_name,
            acl=['ALLOW_NONE']
        )

    def _get_lightos_uuid(self, project_name, volume):
        lightos_name = self._lightos_volname(volume)
        timeout = self.logical_op_timeout

        (status, data) = self._get_lightos_volume(project_name=project_name,
                                                  timeout=timeout,
                                                  vol_name=lightos_name)
        if status != httpstatus.OK or not data:
            LOG.error(
                'Failed to get LightOS volume %s project %s status: \
                %s data: %s',
                lightos_name,
                project_name,
                status,
                str(data))
            raise exception.VolumeNotFound(volume_id=volume)

        lightos_uuid = data.get('UUID')
        if not lightos_uuid:
            LOG.error('Failed to get LightOS volume UUID status: %s, data: %s',
                      status, str(data))
            raise exception.VolumeNotFound(volume_id=volume)

        return lightos_uuid

    def create_volume(self, volume):
        return self._create_volume(volume, src_snapshot_lightos_name=None)

    def create_volume_from_snapshot(self, volume, snapshot):
        snapshotname = self._lightos_snapshotname(snapshot["id"])
        return self._create_volume(volume,
                                   src_snapshot_lightos_name=snapshotname)

    def _create_volume(self, volume, src_snapshot_lightos_name):
        lightos_name = self._lightos_volname(volume)
        project_name = self._get_lightos_project_name(volume)
        lightos_uuid = '<UNKNOWN>'
        vol_state = 'UNKNOWN'

        # first, check if such a volume exists
        # if it exists, we must have created it earlier in a previous
        # invocation of create volume since it takes a while for
        # openstack to retry the call, it's highly unlikely that we created
        # it but it does not show up yet, so assume that if it does not show
        # up, it was never created
        status_code, resp = self._get_lightos_volume(project_name,
                                                     timeout=self.
                                                     logical_op_timeout,
                                                     vol_name=lightos_name)
        if status_code == httpstatus.NOT_FOUND:
            status_code, resp = self._create_new_lightos_volume(
                os_volume=volume,
                project_name=project_name,
                lightos_name=lightos_name,
                src_snapshot_lightos_name=src_snapshot_lightos_name)

        if status_code in (httpstatus.OK, httpstatus.CREATED):
            lightos_uuid = resp['UUID']
            vol_state = self._wait_for_volume_available(
                project_name,
                timeout=self.logical_op_timeout,
                vol_uuid=lightos_uuid)
            allowed_states = ['Available', 'Migrating']
            if vol_state in allowed_states:
                LOG.debug(
                    "LIGHTOS created volume name %s lightos_uuid \
                     %s project %s",
                    lightos_name,
                    lightos_uuid,
                    project_name)
                return

            # if volume was created in failed state we should clean it up
            LOG.warning(
                'LightOS volume with UUID %s project %s last_state is %s',
                lightos_uuid,
                project_name,
                vol_state)
            if vol_state != 'UNKNOWN':
                LOG.debug(
                    'Cleaning up LightOS volume with UUID %s project %s',
                    lightos_uuid,
                    project_name)
                self._delete_lightos_volume(project_name, lightos_uuid)
                # wait for openstack to call us again to create it

        msg = (
            "Did not succeed creating LightOS volume with UUID %(uuid)s"
            " status_code %(code)s last state %(state)s" %
            dict(uuid=lightos_uuid, code=status_code, state=vol_state))
        msg = _(msg)
        raise exception.VolumeBackendAPIException(message=msg)

    def _wait_for_snapshot_available(self, project_name,
                                     timeout,
                                     snapshot_uuid=None,
                                     snapshot_name=None):
        """Wait until the snapshot is available."""
        assert snapshot_uuid or snapshot_name, \
            'LightOS snapshot UUID or name must be supplied'
        # we can stop on any terminal status
        # possible states: Unknown, Creating, Available, Deleting, Deleted,
        # Failed, Updating
        states = ('Available', 'Deleting', 'Deleted', 'Failed', 'UNKNOWN')

        stop = time.time() + timeout
        while time.time() <= stop:
            (status_code,
             resp) = self._get_lightos_snapshot(project_name,
                                                timeout=
                                                self.logical_op_timeout,
                                                snapshot_uuid=snapshot_uuid,
                                                snapshot_name=snapshot_name)
            state = resp.get('state', 'UNKNOWN') if \
                status_code == httpstatus.OK and resp else 'UNKNOWN'
            if state in states and status_code != httpstatus.NOT_FOUND:
                break
            time.sleep(1)

        return state

    def _wait_for_snapshot_deleted(self,
                                   project_name,
                                   timeout,
                                   snapshot_uuid):
        """Wait until the snapshot has been deleted."""
        assert snapshot_uuid, 'LightOS snapshot UUID must be specified'
        states = ('Deleted', 'Deleting', 'UNKNOWN')

        stop = time.time() + timeout
        while time.time() <= stop:
            status_code, resp = (
                self._get_lightos_snapshot(project_name,
                                           timeout=self.logical_op_timeout,
                                           snapshot_uuid=snapshot_uuid))
            if status_code == httpstatus.NOT_FOUND:
                return 'Deleted'
            state = resp.get('state', 'UNKNOWN') if \
                status_code == httpstatus.OK and resp else 'UNKNOWN'
            if state in states:
                break
            time.sleep(1)

        return state

    def _wait_for_volume_deleted(self, project_name, timeout, vol_uuid):
        """Wait until the volume has been deleted."""
        assert vol_uuid, 'LightOS volume UUID must be specified'
        states = ('Deleted', 'Deleting', 'UNKNOWN')

        stop = time.time() + timeout
        while time.time() <= stop:
            (status_code,
             resp) = self._get_lightos_volume(project_name,
                                              timeout=self.logical_op_timeout,
                                              vol_uuid=vol_uuid)
            if status_code == httpstatus.NOT_FOUND:
                return 'Deleted'
            state = resp.get('state', 'UNKNOWN') if \
                status_code == httpstatus.OK and resp else 'UNKNOWN'
            if state in states:
                break
            time.sleep(1)

        return state

    def _delete_lightos_volume(self, project_name, lightos_uuid):
        end = time.time() + self.logical_op_timeout
        while (time.time() < end):
            status_code, resp = (
                self.cluster.send_cmd(
                    cmd='delete_volume',
                    project_name=project_name,
                    timeout=self. logical_op_timeout,
                    volume_uuid=lightos_uuid))
            if status_code == httpstatus.OK:
                break

            LOG.warning(
                "delete_volume for volume with LightOS UUID %s failed \
                with status code %s response %s",
                lightos_uuid,
                status_code,
                resp)
            time.sleep(1)
        else:  # no break
            LOG.error(
                "Failed to delete volume with LightOS UUID %s. Final status \
                code %s response %s",
                lightos_uuid,
                status_code,
                resp)
            return False

        deleted_state = self._wait_for_volume_deleted(
            project_name, timeout=self.logical_op_timeout,
            vol_uuid=lightos_uuid)
        return deleted_state in ('Deleted', 'Deleting', 'UNKNOWN')

    def delete_volume(self, volume):
        """Delete volume."""
        project_name = self._get_lightos_project_name(volume)
        try:
            lightos_uuid = self._get_lightos_uuid(project_name, volume)
        except exception.VolumeNotFound:
            return True

        if not self._delete_lightos_volume(project_name, lightos_uuid):
            msg = ('Failed to delete LightOS volume with UUID'
                   ' %(uuid)s project %(project_name)s' % (
                       dict(uuid=lightos_uuid, project_name=project_name)))
            raise exception.VolumeBackendAPIException(message=msg)

    def get_vol_by_id(self, volume):
        LOG.warning('UNIMPLEMENTED: get vol by id')

    def get_vols(self):
        LOG.warning('UNIMPLEMENTED: get vols')

    def check_for_setup_error(self):
        subsysnqn = self.cluster.subsystemNQN
        if not subsysnqn:
            msg = ('LIGHTOS: Cinder driver requires the'
                   ' LightOS cluster subsysnqn')
            raise exception.VolumeBackendAPIException(message=msg)

        hostnqn = (
            self.connector.get_connector_properties(
                utils.get_root_helper())['nqn'])
        if not hostnqn:
            msg = ("LIGHTOS: Cinder driver requires a local hostnqn for"
                   " image_to/from_volume operations")
            raise exception.VolumeBackendAPIException(message=msg)

    def get_cluster_info(self):
        status_code, cluster_info = self.cluster.send_cmd(
            cmd='get_cluster_info', timeout=self.logical_op_timeout)
        if status_code == httpstatus.UNAUTHORIZED:
            msg = f'LIGHTOS: failed to connect to cluster. code: {status_code}'
            raise exception.InvalidAuthKey(message=_(msg))
        if status_code != httpstatus.OK:
            msg = 'LIGHTOS: Could not connect to LightOS cluster'
            raise exception.VolumeBackendAPIException(message=_(msg))

        LOG.info("Connected to LightOS cluster %s subsysnqn %s",
                 cluster_info['UUID'], cluster_info['subsystemNQN'])
        self.cluster.lightos_cluster_uuid = cluster_info['UUID']
        self.cluster.subsystemNQN = cluster_info['subsystemNQN']

    def get_cluster_stats(self):
        status_code, cluster_info = self.cluster.send_cmd(
            cmd='get_cluster', timeout=self.logical_op_timeout)
        if status_code != httpstatus.OK:
            msg = 'LIGHTOS: Could not connect to LightOS cluster'
            raise exception.VolumeBackendAPIException(message=_(msg))

        return cluster_info['statistics']

    def valid_nodes_info(self, nodes_info):
        if not nodes_info or 'nodes' not in nodes_info:
            return False

        return True

    def wait_for_lightos_cluster(self):
        cmd = 'get_nodes'
        end = time.time() + self.logical_op_timeout
        while (time.time() < end):
            status_code, nodes_info = self.cluster.send_cmd(
                cmd=cmd, timeout=self.logical_op_timeout)
            if status_code != httpstatus.OK or not self.valid_nodes_info(
                    nodes_info):
                time.sleep(1)
                continue

            return nodes_info

        # bail out if we got here, timeout elapsed
        msg = 'Failed to get nodes, last status was {} nodes_info {}'.format(
            status_code, nodes_info)
        raise exception.VolumeBackendAPIException(message=_(msg))

    def do_setup(self, context):

        self.get_cluster_info()
        nodes_info = self.wait_for_lightos_cluster()

        self.cluster.targets = dict()
        node_list = nodes_info['nodes']
        for node in node_list:
            self.cluster.targets[node['UUID']] = node

        # reduce the logical op timeout if single server LightOS cluster
        if len(node_list) == 1:
            self.logical_op_timeout = self.configuration. \
                lightos_api_service_timeout + 10

    def extend_volume(self, volume, size):
        # loop because lightos api is async
        end = time.time() + self.logical_op_timeout
        while (time.time() < end):
            try:
                finished = self._extend_volume(volume, size)
                if finished:
                    break
            except exception.VolumeNotFound as e:
                raise e
            except Exception as e:
                # bail out if the time out elapsed...
                if time.time() >= end:
                    LOG.warning('Timed out extend volume operation')
                    raise e
                # if we still have more time, just print the exception
                LOG.warning(
                    'caught this in extend_volume() ... will retry: %s',
                    str(e))
                time.sleep(1)

    def _extend_volume(self, volume, size):
        lightos_volname = self._lightos_volname(volume)
        project_name = self._get_lightos_project_name(volume)

        try:
            (status, data) = self._get_lightos_volume(
                project_name,
                timeout=self.
                logical_op_timeout,
                vol_name=lightos_volname)
            if status != httpstatus.OK or not data:
                LOG.error(
                    'Failed to get LightOS volume status: %s data: %s',
                    status,
                    str(data))
                raise exception.VolumeNotFound(volume_id=volume.id)
            lightos_uuid = data['UUID']
            etag = data.get('ETag', '')
        except Exception as e:
            raise e

        try:
            code, message = self.cluster.send_cmd(
                cmd='extend_volume',
                project_name=project_name,
                timeout=self.logical_op_timeout,
                volume_uuid=lightos_uuid,
                size=str(size) + ' gib',
                etag=etag
            )
            if code == httpstatus.OK:
                LOG.info(
                    "Successfully extended volume %s project %s size:%s",
                    volume,
                    project_name,
                    size)
            else:
                raise exception.ExtendVolumeError(reason=message)

        except exception.ExtendVolumeError as e:
            raise e
        except Exception as e:
            raise exception.ExtendVolumeError(raised_exception=e)
        return True

    @staticmethod
    def byte_to_gb(bbytes):
        return int(int(bbytes) / units.Gi)

    def get_volume_stats(self, refresh=False):
        """Retrieve stats info for the volume *service*,

        not a specific volume.
        """

        LOG.debug("getting volume stats (refresh=%s)", refresh)

        if not refresh:
            return self._stats

        backend_name = self.configuration.safe_get('volume_backend_name')
        res_percentage = self.configuration.safe_get('reserved_percentage')
        # as a tenant we dont have access to cluster stats
        # in the future we might expose this per project via get_project API
        # currently we remove this stats call.
        # cluster_stats = self.get_cluster_stats()

        data = {'vendor_name': 'LightOS Storage',
                'volume_backend_name': backend_name or self.__class__.__name__,
                'driver_version': self.VERSION,
                'storage_protocol': constants.LIGHTOS,
                'reserved_percentage': res_percentage,
                'QoS_support': False,
                'online_extend_support': True,
                'thin_provisioning_support': True,
                'compression': [True, False],
                'multiattach': True}
        # data['total_capacity_gb'] =
        # self.byte_to_gb(cluster_stats['effectivePhysicalStorage'])
        # It would be preferable to return
        # self.byte_to_gb(cluster_stats['freePhysicalStorage'])
        # here but we return 'infinite' due to the Cinder bug described in
        # https://bugs.launchpad.net/cinder/+bug/1871371
        data['free_capacity_gb'] = 'infinite'
        self._stats = data

        return self._stats

    def _get_connection_properties(self, project_name, volume):
        lightos_targets = {}
        for target in self.cluster.targets.values():
            properties = dict()
            ep = urlparse('//' + target['nvmeEndpoint'])
            properties['target_portal'] = ep.hostname
            properties['target_port'] = 8009  # spec specified discovery port
            properties['transport_type'] = 'tcp'
            lightos_targets[ep.hostname] = properties

        server_properties = {}
        server_properties['lightos_nodes'] = lightos_targets
        server_properties['uuid'] = (
            self._get_lightos_uuid(project_name, volume))
        server_properties['subsysnqn'] = self.cluster.subsystemNQN

        return server_properties

    def set_volume_acl(self, project_name, lightos_uuid, acl, etag):
        return self.cluster.send_cmd(
            cmd='update_volume',
            project_name=project_name,
            timeout=self.logical_op_timeout,
            volume_uuid=lightos_uuid,
            acl=acl,
            etag=etag
        )

    def __add_volume_acl(self, project_name, lightos_volname, acl_to_add):
        (status, data) = self._get_lightos_volume(project_name,
                                                  self.logical_op_timeout,
                                                  vol_name=lightos_volname)
        if status != httpstatus.OK or not data:
            LOG.error('Failed to get LightOS volume %s status %s data %s',
                      lightos_volname, status, data)
            return False

        lightos_uuid = data.get('UUID')
        if not lightos_uuid:
            LOG.warning('Got LightOS volume without UUID?! data: %s', data)
            return False

        acl = data.get('acl')
        if not acl:
            LOG.warning('Got LightOS volume without ACL?! data: %s', data)
            return False

        acl = acl.get('values', [])

        # remove ALLOW_NONE and add our acl_to_add if not already there
        if 'ALLOW_NONE' in acl:
            acl.remove('ALLOW_NONE')
        if acl_to_add not in acl:
            acl.append(acl_to_add)

        return self.set_volume_acl(
            project_name,
            lightos_uuid,
            acl,
            etag=data.get(
                'ETag',
                ''))

    def add_volume_acl(self, project_name, volume, acl_to_add):
        LOG.debug(
            'add_volume_acl got volume %s project %s acl %s',
            volume,
            project_name,
            acl_to_add)
        lightos_volname = self._lightos_volname(volume)
        return self.update_volume_acl(
            self.__add_volume_acl,
            project_name,
            lightos_volname,
            acl_to_add)

    def __remove_volume_acl(
            self,
            project_name,
            lightos_volname,
            acl_to_remove):
        (status, data) = self._get_lightos_volume(project_name,
                                                  self.logical_op_timeout,
                                                  vol_name=lightos_volname)
        if not data:
            LOG.error(
                'Could not get data for LightOS volume %s project %s',
                lightos_volname,
                project_name)
            return False

        lightos_uuid = data.get('UUID')
        if not lightos_uuid:
            LOG.warning('Got LightOS volume without UUID?! data: %s', data)
            return False

        acl = data.get('acl')
        if not acl:
            LOG.warning('Got LightOS volume without ACL?! data: %s', data)
            return False

        acl = acl.get('values')
        if not acl:
            LOG.warning(
                'Got LightOS volume without ACL values?! data: %s', data)
            return False

        try:
            acl.remove(acl_to_remove)
        except ValueError:
            LOG.warning(
                'Could not remove acl %s from LightOS volume %s project \
                %s with acl %s',
                acl_to_remove,
                lightos_volname,
                project_name,
                acl)

        # if the ACL is empty here, put in ALLOW_NONE
        if not acl:
            acl.append('ALLOW_NONE')

        return self.set_volume_acl(
            project_name,
            lightos_uuid,
            acl,
            etag=data.get(
                'ETag',
                ''))

    def __overwrite_volume_acl(
            self,
            project_name,
            lightos_volname,
            acl):
        status, data = self._get_lightos_volume(project_name,
                                                self.logical_op_timeout,
                                                vol_name=lightos_volname)
        if not data:
            LOG.error(
                'Could not get data for LightOS volume %s project %s',
                lightos_volname,
                project_name)
            return False

        lightos_uuid = data.get('UUID')
        if not lightos_uuid:
            LOG.warning('Got LightOS volume without UUID?! data: %s', data)
            return False

        return self.set_volume_acl(
            project_name,
            lightos_uuid,
            acl,
            etag=data.get(
                'ETag',
                ''))

    def remove_volume_acl(self, project_name, volume, acl_to_remove):
        lightos_volname = self._lightos_volname(volume)
        LOG.debug('remove_volume_acl volume %s project %s acl %s',
                  volume, project_name, acl_to_remove)
        return self.update_volume_acl(
            self.__remove_volume_acl,
            project_name,
            lightos_volname,
            acl_to_remove)

    def remove_all_volume_acls(self, project_name, volume):
        lightos_volname = self._lightos_volname(volume)
        LOG.debug('remove_all_volume_acls volume %s project %s',
                  volume, project_name)
        return self.update_volume_acl(
            self.__overwrite_volume_acl,
            project_name,
            lightos_volname,
            ['ALLOW_NONE'])

    def update_volume_acl(self, func, project_name, lightos_volname, acl):
        # loop because lightos api is async
        end = time.time() + self.logical_op_timeout
        first_iteration = True
        while (time.time() < end):
            if not first_iteration:
                time.sleep(1)
            first_iteration = False
            res = func(project_name, lightos_volname, acl)
            if not isinstance(res, tuple):
                LOG.debug('Update_volume: func %s(%s project %s) failed',
                          func, lightos_volname, project_name)
                continue
            if len(res) != 2:
                LOG.debug("Unexpected number of values to unpack")
                continue
            (status, resp) = res
            if status != httpstatus.OK:
                LOG.debug(
                    'update_volume: func %s(%s project %s) got \
                    http status %s',
                    func,
                    lightos_volname,
                    project_name,
                    status)
            else:
                break

        # bail out if the time out elapsed...
        if time.time() >= end:
            LOG.warning(
                'Timed out %s(%s project %s)',
                func,
                lightos_volname,
                project_name)
            return False

        # or the call succeeded and we need to wait
        # for the volume to stabilize
        vol_state = self._wait_for_volume_available(
            project_name, timeout=end - time.time(), vol_name=lightos_volname)
        allowed_states = ['Available', 'Migrating']
        if vol_state not in allowed_states:
            LOG.warning(
                'Timed out waiting for volume %s project %s to stabilize, \
                last state %s',
                lightos_volname,
                project_name,
                vol_state)
            return False

        return True

    def _wait_for_volume_acl(
            self,
            project_name,
            lightos_volname,
            acl,
            requested_membership):
        end = time.time() + self.logical_op_timeout
        while (time.time() < end):
            (status, resp) = self._get_lightos_volume(
                project_name,
                self.logical_op_timeout,
                vol_name=lightos_volname)
            if status == httpstatus.OK:
                if not resp or not resp.get('acl'):
                    LOG.warning(
                        'Got LightOS volume %s without ACL?! data: %s',
                        lightos_volname,
                        resp)
                    return False

                volume_acls = resp.get('acl').get('values', [])
                membership = acl in volume_acls
                if membership == requested_membership:
                    return True

            LOG.debug(
                'ACL did not settle for volume %s project %s, status \
                %s resp %s',
                lightos_volname,
                project_name,
                status,
                resp)
            time.sleep(1)
        LOG.warning(
            'ACL did not settle for volume %s, giving up',
            lightos_volname)
        return False

    def create_snapshot(self, snapshot):
        snapshot_name = self._lightos_snapshotname(snapshot["id"])
        src_volume_name = self._lightos_volname(snapshot["volume"])
        project_name = self._get_lightos_project_name(snapshot.volume)
        self._create_snapshot(project_name, snapshot_name, src_volume_name)

    @coordination.synchronized('lightos-create_snapshot-{src_volume_name}')
    def _create_snapshot(self, project_name, snapshot_name, src_volume_name):
        (status_code_get, response) = self._get_lightos_snapshot(
            project_name, self.logical_op_timeout,
            snapshot_name=snapshot_name)
        if status_code_get != httpstatus.OK:
            end = time.time() + self.logical_op_timeout
            while (time.time() < end):
                (status_code_create, response) = self.cluster.send_cmd(
                    cmd='create_snapshot',
                    project_name=project_name,
                    timeout=self.logical_op_timeout,
                    name=snapshot_name,
                    src_volume_name=src_volume_name,
                )

                if status_code_create == httpstatus.INTERNAL_SERVER_ERROR:
                    pass
                else:
                    break

                time.sleep(1)

            if status_code_create != httpstatus.OK:
                msg = ('Did not succeed creating LightOS snapshot %s'
                       ' project %s'
                       ' status code %s response %s' %
                       (snapshot_name, project_name, status_code_create,
                        response))
                raise exception.VolumeBackendAPIException(message=_(msg))

        state = self._wait_for_snapshot_available(project_name,
                                                  timeout=
                                                  self.logical_op_timeout,
                                                  snapshot_name=snapshot_name)

        if state == 'Available':
            LOG.debug(
                'Successfully created LightOS snapshot %s', snapshot_name)
            return

        LOG.error(
            'Failed to create snapshot %s project %s for volume %s. \
            state = %s.',
            snapshot_name,
            project_name,
            src_volume_name,
            state)
        try:
            self._delete_lightos_snapshot(project_name, snapshot_name)
        except exception.CinderException as ex:
            LOG.warning("Error deleting snapshot during cleanup: %s", ex)

        msg = ('Did not succeed creating LightOS snapshot %s project'
               '%s last state %s' % (snapshot_name, project_name, state))
        raise exception.VolumeBackendAPIException(message=_(msg))

    def delete_snapshot(self, snapshot):
        lightos_snapshot_name = self._lightos_snapshotname(snapshot["id"])
        project_name = self._get_lightos_project_name(snapshot.volume)
        self._delete_lightos_snapshot(project_name=project_name,
                                      snapshot_name=lightos_snapshot_name)

    def _get_lightos_snapshot_uuid(self, project_name, lightos_snapshot_name):
        (status_code, data) = self._get_lightos_snapshot(
            project_name=project_name,
            timeout=self.logical_op_timeout,
            snapshot_name=lightos_snapshot_name)

        if status_code == httpstatus.OK:
            uuid = data.get("UUID")
            if uuid:
                return uuid

        if status_code == httpstatus.NOT_FOUND:
            return None

        msg = ('Unable to fetch UUID of snapshot named %s. status code'
               ' %s data %s' % (lightos_snapshot_name, status_code, data))
        raise exception.VolumeBackendAPIException(message=_(msg))

    def _delete_lightos_snapshot(self, project_name, snapshot_name):
        snapshot_uuid = self._get_lightos_snapshot_uuid(
            project_name, snapshot_name)
        if snapshot_uuid is None:
            LOG.warning(
                "Unable to find lightos snapshot %s project %s for deletion",
                snapshot_name,
                project_name)
            return False

        (status_code, _) = self.cluster.send_cmd(cmd='delete_snapshot',
                                                 project_name=project_name,
                                                 timeout=self.
                                                 logical_op_timeout,
                                                 snapshot_uuid=snapshot_uuid)
        if status_code == httpstatus.OK:
            state = self._wait_for_snapshot_deleted(
                project_name,
                timeout=self.logical_op_timeout,
                snapshot_uuid=snapshot_uuid)
            if state in ('Deleted', 'Deleting', 'UNKNOWN'):
                LOG.debug(
                    "Successfully detected that snapshot %s was deleted.",
                    snapshot_name)
                return True
            LOG.warning("Snapshot %s was not deleted. It is in state %s.",
                        snapshot_name, state)
            return False
        LOG.warning(
            "Request to delete snapshot %s"
            " was rejected with status code %s.",
            snapshot_name,
            status_code)
        return False

    def initialize_connection(self, volume, connector):
        hostnqn = connector.get('nqn')
        found_dsc = connector.get('found_dsc')
        LOG.debug(
            'initialize_connection: connector hostnqn is %s found_dsc %s',
            hostnqn,
            found_dsc)
        if not hostnqn:
            msg = 'Connector (%s) did not contain a hostnqn, aborting' % (
                connector)
            raise exception.VolumeBackendAPIException(message=_(msg))

        if not found_dsc:
            msg = ('Connector (%s) did not indicate a discovery'
                   'client, aborting' % (connector))
            raise exception.VolumeBackendAPIException(message=_(msg))

        lightos_volname = self._lightos_volname(volume)
        project_name = self._get_lightos_project_name(volume)
        success = self.add_volume_acl(project_name, volume, hostnqn)
        if not success or not self._wait_for_volume_acl(
                project_name, lightos_volname, hostnqn, True):
            msg = ('Could not add ACL for hostnqn %s LightOS volume'
                   ' %s, aborting' % (hostnqn, lightos_volname))
            raise exception.VolumeBackendAPIException(message=_(msg))

        props = self._get_connection_properties(project_name, volume)
        return {'driver_volume_type': ('lightos'), 'data': props}

    def terminate_connection(self, volume, connector, **kwargs):
        force = 'force' in kwargs
        hostnqn = connector.get('nqn') if connector else None
        LOG.debug(
            'terminate_connection: force %s kwargs %s hostnqn %s',
            force,
            kwargs,
            hostnqn)

        project_name = self._get_lightos_project_name(volume)

        if not hostnqn:
            if force:
                LOG.debug(
                    'Terminating connection with extreme prejudice for \
                    volume %s',
                    volume)
                self.remove_all_volume_acls(project_name, volume)
                return

            msg = 'Connector (%s) did not return a hostnqn, aborting' % (
                connector)
            raise exception.VolumeBackendAPIException(message=_(msg))

        lightos_volname = self._lightos_volname(volume)
        project_name = self._get_lightos_project_name(volume)
        success = self.remove_volume_acl(project_name, volume, hostnqn)
        if not success or not self._wait_for_volume_acl(
                project_name, lightos_volname, hostnqn, False):
            LOG.warning(
                'Could not remove ACL for hostnqn %s LightOS \
                volume %s, limping along',
                hostnqn,
                lightos_volname)

    def _init_vendor_properties(self):
        # compression is one of the standard properties,
        # no need to add it here
        # see the definition of this function in cinder/volume/driver.py
        properties = {}
        self._set_property(
            properties,
            "lightos:num_replicas",
            "Number of replicas for LightOS volume",
            _(
                "Specifies the number of replicas to create for the \
                LightOS volume."),
            "integer",
            minimum=1,
            maximun=3,
            default=3)

        return properties, 'lightos'

    def backup_use_temp_snapshot(self):
        return False

    def snapshot_revert_use_temp_snapshot(self):
        """Disable the use of a temporary snapshot on revert."""
        return False

    def snapshot_remote_attachable(self):
        """LightOS does not support 'mount a snapshot'"""
        return False



# Copyright 2013 IBM Corp.
# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
from unittest import mock

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.ibm.ibm_storage import ibm_storage
from cinder.volume import volume_types

FAKE = "fake"
FAKE2 = "fake2"
CANNOT_DELETE = "Can not delete"
TOO_BIG_VOLUME_SIZE = 12000
POOL_SIZE = 100
GROUP_ID = 1
VOLUME = {'size': 16,
          'name': FAKE,
          'id': 1,
          'status': 'available'}
VOLUME2 = {'size': 32,
           'name': FAKE2,
           'id': 2,
           'status': 'available'}
GROUP_VOLUME = {'size': 16,
                'name': FAKE,
                'id': 3,
                'group_id': GROUP_ID,
                'status': 'available'}

MANAGED_FAKE = "managed_fake"
MANAGED_VOLUME = {'size': 16,
                  'name': MANAGED_FAKE,
                  'id': 2}

REPLICA_FAKE = "repicated_fake"
REPLICATED_VOLUME = {'size': 64,
                     'name': REPLICA_FAKE,
                     'id': '2',
                     'replication_status': fields.ReplicationStatus.ENABLED}

REPLICATED_VOLUME_DISABLED = REPLICATED_VOLUME.copy()
REPLICATED_VOLUME_DISABLED['replication_status'] = (
    fields.ReplicationStatus.DISABLED)

REPLICATION_TARGETS = [{'target_device_id': 'fakedevice'}]
SECONDARY = 'fakedevice'
FAKE_FAILOVER_HOST = 'fakehost@fakebackend#fakepool'
FAKE_PROVIDER_LOCATION = 'fake_provider_location'
FAKE_DRIVER_DATA = 'fake_driver_data'

CONTEXT = {}

FAKESNAPSHOT = 'fakesnapshot'
SNAPSHOT = {'name': 'fakesnapshot',
            'id': 3}

GROUP = {'id': GROUP_ID, }
GROUP_SNAPSHOT_ID = 1
GROUP_SNAPSHOT = {'id': GROUP_SNAPSHOT_ID,
                  'group_id': GROUP_ID}

CONNECTOR = {'initiator': "iqn.2012-07.org.fake:01:948f189c4695", }

FAKE_PROXY = 'cinder.tests.unit.volume.drivers.ibm.test_ibm_storage' \
    '.IBMStorageFakeProxyDriver'


class IBMStorageFakeProxyDriver(object):
    """Fake IBM Storage driver

    Fake IBM Storage driver for IBM XIV, Spectrum Accelerate,
    FlashSystem A9000, FlashSystem A9000R and DS8000 storage systems.
    """

    def __init__(self, ibm_storage_info, logger, expt,
                 driver=None, active_backend_id=None, host=None):
        """Initialize Proxy."""

        self.ibm_storage_info = ibm_storage_info
        self.logger = logger
        self.exception = expt
        self.storage_portal = \
            self.storage_iqn = FAKE

        self.volumes = {}
        self.snapshots = {}
        self.driver = driver

    def setup(self, context):
        if self.ibm_storage_info['user'] != self.driver\
                .configuration.san_login:
            raise self.exception.NotAuthorized()

        if self.ibm_storage_info['address'] != self.driver\
                .configuration.san_ip:
            raise self.exception.HostNotFound(host='fake')

    def create_volume(self, volume):
        if volume['size'] > POOL_SIZE:
            raise self.exception.VolumeBackendAPIException(data='blah')
        self.volumes[volume['name']] = volume

    def volume_exists(self, volume):
        return self.volumes.get(volume['name'], None) is not None

    def delete_volume(self, volume):
        if self.volumes.get(volume['name'], None) is not None:
            del self.volumes[volume['name']]

    def manage_volume_get_size(self, volume, existing_ref):
        if self.volumes.get(existing_ref['source-name'], None) is None:
            raise self.exception.VolumeNotFound(volume_id=volume['id'])
        return self.volumes[existing_ref['source-name']]['size']

    def manage_volume(self, volume, existing_ref):
        if self.volumes.get(existing_ref['source-name'], None) is None:
            raise self.exception.VolumeNotFound(volume_id=volume['id'])
        volume['size'] = MANAGED_VOLUME['size']
        return {}

    def unmanage_volume(self, volume):
        pass

    def initialize_connection(self, volume, connector):
        if not self.volume_exists(volume):
            raise self.exception.VolumeNotFound(volume_id=volume['id'])
        lun_id = volume['id']

        self.volumes[volume['name']]['attached'] = connector

        return {'driver_volume_type': 'iscsi',
                'data': {'target_discovered': True,
                         'target_portal': self.storage_portal,
                         'target_iqn': self.storage_iqn,
                         'target_lun': lun_id,
                         'volume_id': volume['id'],
                         'multipath': True,
                         'provider_location': "%s,1 %s %s" % (
                             self.storage_portal,
                             self.storage_iqn,
                             lun_id), },
                }

    def terminate_connection(self, volume, connector):
        if not self.volume_exists(volume):
            raise self.exception.VolumeNotFound(volume_id=volume['id'])
        if not self.is_volume_attached(volume, connector):
            raise self.exception.NotFound(_('Volume not found for '
                                            'instance %(instance_id)s.')
                                          % {'instance_id': 'fake'})
        del self.volumes[volume['name']]['attached']

    def is_volume_attached(self, volume, connector):
        if not self.volume_exists(volume):
            raise self.exception.VolumeNotFound(volume_id=volume['id'])

        return (self.volumes[volume['name']].get('attached', None)
                == connector)

    def get_replication_status(self, context, volume):
        if volume['replication_status'] == 'invalid_status_val':
            raise exception.CinderException()
        return {'replication_status': 'active'}

    def retype(self, ctxt, volume, new_type, diff, host):
        volume['easytier'] = new_type['extra_specs']['easytier']
        return True, volume

    def create_group(self, ctxt, group):

        volumes = [volume for k, volume in self.volumes.items()
                   if volume['group_id'] == group['id']]

        if volumes:
            raise exception.CinderException(
                message='The group id of volume may be wrong.')

        return {'status': fields.GroupStatus.AVAILABLE}

    def delete_group(self, ctxt, group, volumes):
        for volume in self.volumes.values():
            if group.get('id') == volume.get('group_id'):
                if volume['name'] == CANNOT_DELETE:
                    raise exception.VolumeBackendAPIException(
                        message='Volume can not be deleted')
                else:
                    volume['status'] = 'deleted'
                    volumes.append(volume)

        # Delete snapshots in group
        self.snapshots = {k: snap for k, snap in self.snapshots.items()
                          if not(snap.get('group_id') == group.get('id'))}

        # Delete volume in group
        self.volumes = {k: vol for k, vol in self.volumes.items()
                        if not(vol.get('group_id') == group.get('id'))}

        return {'status': fields.GroupStatus.DELETED}, volumes

    def update_group(self, context, group, add_volumes, remove_volumes):
        model_update = {'status': fields.GroupStatus.AVAILABLE}
        return model_update, None, None

    def create_group_from_src(self, context, group, volumes, group_snapshot,
                              snapshots, source_group=None, source_vols=None):

        return None, None

    def create_group_snapshot(self, ctxt, group_snapshot, snapshots):
        for volume in self.volumes.values():
            if group_snapshot.get('group_id') == volume.get('group_id'):
                if volume['size'] > POOL_SIZE / 2:
                    raise self.exception.VolumeBackendAPIException(data='blah')

                snapshot = copy.deepcopy(volume)
                snapshot['name'] = (
                    CANNOT_DELETE if snapshot['name'] == CANNOT_DELETE
                    else snapshot['name'] + 'Snapshot')
                snapshot['status'] = 'available'
                snapshot['group_snapshot_id'] = group_snapshot.get('id')
                snapshot['group_id'] = group_snapshot.get('group_id')
                self.snapshots[snapshot['name']] = snapshot
                snapshots.append(snapshot)

        return {'status': fields.GroupSnapshotStatus.AVAILABLE}, snapshots

    def delete_group_snapshot(self, ctxt, group_snapshot, snapshots):
        updated_snapshots = []
        for snapshot in snapshots:
            if snapshot['name'] == CANNOT_DELETE:
                raise exception.VolumeBackendAPIException(
                    message='Snapshot can not be deleted')
            else:
                snapshot['status'] = 'deleted'
                updated_snapshots.append(snapshot)

        # Delete snapshots in group
        self.snapshots = {k: snap for k, snap in self.snapshots.items()
                          if not(snap.get('group_id')
                                 == group_snapshot.get('group_snapshot_id'))}

        return {'status': 'deleted'}, updated_snapshots

    def freeze_backend(self, context):
        return True

    def thaw_backend(self, context):
        return True

    def failover_host(self, context, volumes, secondary_id, groups=None):
        target_id = 'BLA'
        volume_update_list = []
        for volume in volumes:
            status = 'failed-over'
            if volume['replication_status'] == 'invalid_status_val':
                status = 'error'
            volume_update_list.append(
                {'volume_id': volume['id'],
                 'updates': {'replication_status': status}})

        return target_id, volume_update_list, []

    def enable_replication(self, context, group, volumes):
        vol_status = []
        for vol in volumes:
            vol_status.append(
                {'id': vol['id'],
                 'replication_status': fields.ReplicationStatus.ENABLED})
        return (
            {'replication_status': fields.ReplicationStatus.ENABLED},
            vol_status)

    def disable_replication(self, context, group, volumes):
        volume_update_list = []
        for volume in volumes:
            volume_update_list.append(
                {'id': volume['id'],
                 'replication_status': fields.ReplicationStatus.DISABLED})
        return (
            {'replication_status': fields.ReplicationStatus.DISABLED},
            volume_update_list)

    def failover_replication(self, context, group, volumes, secondary_id):
        volume_update_list = []
        for volume in volumes:
            volume_update_list.append(
                {'id': volume['id'],
                 'replication_status': fields.ReplicationStatus.FAILED_OVER})
        return ({'replication_status': fields.ReplicationStatus.FAILED_OVER},
                volume_update_list)

    def get_replication_error_status(self, context, groups):
        return(
            [{'group_id': groups[0]['id'],
              'replication_status': fields.ReplicationStatus.ERROR}],
            [{'volume_id': VOLUME['id'],
              'replication_status': fields.ReplicationStatus.ERROR}])


class IBMStorageVolumeDriverTest(test.TestCase):
    """Test IBM Storage driver

    Test IBM Storage driver for IBM XIV, Spectrum Accelerate,
    FlashSystem A9000, FlashSystem A9000R and DS8000 storage Systems.
    """

    def setUp(self):
        """Initialize IBM Storage Driver."""
        super(IBMStorageVolumeDriverTest, self).setUp()

        configuration = mock.Mock(conf.Configuration)
        configuration.san_is_local = False
        configuration.proxy = FAKE_PROXY
        configuration.connection_type = 'iscsi'
        configuration.chap = 'disabled'
        configuration.san_ip = FAKE
        configuration.management_ips = FAKE
        configuration.san_login = FAKE
        configuration.san_clustername = FAKE
        configuration.san_password = FAKE
        configuration.append_config_values(mock.ANY)

        self.driver = ibm_storage.IBMStorageDriver(
            configuration=configuration)

    def test_initialized_should_set_ibm_storage_info(self):
        """Test that the san flags are passed to the IBM proxy."""

        self.assertEqual(
            self.driver.proxy.ibm_storage_info['user'],
            self.driver.configuration.san_login)
        self.assertEqual(
            self.driver.proxy.ibm_storage_info['password'],
            self.driver.configuration.san_password)
        self.assertEqual(
            self.driver.proxy.ibm_storage_info['address'],
            self.driver.configuration.san_ip)
        self.assertEqual(
            self.driver.proxy.ibm_storage_info['vol_pool'],
            self.driver.configuration.san_clustername)

    def test_setup_should_fail_if_credentials_are_invalid(self):
        """Test that the proxy validates credentials."""

        self.driver.proxy.ibm_storage_info['user'] = 'invalid'
        self.assertRaises(exception.NotAuthorized, self.driver.do_setup, None)

    def test_setup_should_fail_if_connection_is_invalid(self):
        """Test that the proxy validates connection."""

        self.driver.proxy.ibm_storage_info['address'] = \
            'invalid'
        self.assertRaises(exception.HostNotFound, self.driver.do_setup, None)

    def test_create_volume(self):
        """Test creating a volume."""

        self.driver.do_setup(None)
        self.driver.create_volume(VOLUME)
        has_volume = self.driver.proxy.volume_exists(VOLUME)
        self.assertTrue(has_volume)
        self.driver.delete_volume(VOLUME)

    def test_volume_exists(self):
        """Test the volume exist method with a volume that doesn't exist."""

        self.driver.do_setup(None)

        self.assertFalse(
            self.driver.proxy.volume_exists({'name': FAKE})
        )

    def test_delete_volume(self):
        """Verify that a volume is deleted."""

        self.driver.do_setup(None)
        self.driver.create_volume(VOLUME)
        self.driver.delete_volume(VOLUME)
        has_volume = self.driver.proxy.volume_exists(VOLUME)
        self.assertFalse(has_volume)

    def test_delete_volume_should_fail_for_not_existing_volume(self):
        """Verify that deleting a non-existing volume is OK."""

        self.driver.do_setup(None)
        self.driver.delete_volume(VOLUME)

    def test_create_volume_should_fail_if_no_pool_space_left(self):
        """Verify that the proxy validates volume pool space."""

        self.driver.do_setup(None)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume,
                          {'name': FAKE,
                           'id': 1,
                           'size': TOO_BIG_VOLUME_SIZE})

    def test_initialize_connection(self):
        """Test that inititialize connection attaches volume to host."""

        self.driver.do_setup(None)
        self.driver.create_volume(VOLUME)
        self.driver.initialize_connection(VOLUME, CONNECTOR)

        self.assertTrue(
            self.driver.proxy.is_volume_attached(VOLUME, CONNECTOR))

        self.driver.terminate_connection(VOLUME, CONNECTOR)
        self.driver.delete_volume(VOLUME)

    def test_initialize_connection_should_fail_for_non_existing_volume(self):
        """Verify that initialize won't work for non-existing volume."""

        self.driver.do_setup(None)
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.initialize_connection,
                          VOLUME,
                          CONNECTOR)

    def test_terminate_connection(self):
        """Test terminating a connection."""

        self.driver.do_setup(None)
        self.driver.create_volume(VOLUME)
        self.driver.initialize_connection(VOLUME, CONNECTOR)
        self.driver.terminate_connection(VOLUME, CONNECTOR)

        self.assertFalse(self.driver.proxy.is_volume_attached(
            VOLUME,
            CONNECTOR))

        self.driver.delete_volume(VOLUME)

    def test_terminate_connection_should_fail_on_non_existing_volume(self):
        """Test that terminate won't work for non-existing volumes."""

        self.driver.do_setup(None)
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.terminate_connection,
                          VOLUME,
                          CONNECTOR)

    def test_manage_existing_get_size(self):
        """Test that manage_existing_get_size returns the expected size. """

        self.driver.do_setup(None)
        self.driver.create_volume(MANAGED_VOLUME)
        existing_ref = {'source-name': MANAGED_VOLUME['name']}
        return_size = self.driver.manage_existing_get_size(
            VOLUME,
            existing_ref)
        self.assertEqual(MANAGED_VOLUME['size'], return_size)

        # cover both case, whether driver renames the volume or not
        self.driver.delete_volume(VOLUME)
        self.driver.delete_volume(MANAGED_VOLUME)

    def test_manage_existing_get_size_should_fail_on_non_existing_volume(self):
        """Test that manage_existing_get_size fails on non existing volume. """

        self.driver.do_setup(None)
        # on purpose - do NOT create managed volume
        existing_ref = {'source-name': MANAGED_VOLUME['name']}
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.manage_existing_get_size,
                          VOLUME,
                          existing_ref)

    def test_manage_existing(self):
        """Test that manage_existing returns successfully. """

        self.driver.do_setup(None)
        self.driver.create_volume(MANAGED_VOLUME)
        existing_ref = {'source-name': MANAGED_VOLUME['name']}
        self.driver.manage_existing(VOLUME, existing_ref)
        self.assertEqual(MANAGED_VOLUME['size'], VOLUME['size'])

        # cover both case, whether driver renames the volume or not
        self.driver.delete_volume(VOLUME)
        self.driver.delete_volume(MANAGED_VOLUME)

    def test_manage_existing_should_fail_on_non_existing_volume(self):
        """Test that manage_existing fails on non existing volume. """

        self.driver.do_setup(None)
        # on purpose - do NOT create managed volume
        existing_ref = {'source-name': MANAGED_VOLUME['name']}
        self.assertRaises(exception.VolumeNotFound,
                          self.driver.manage_existing,
                          VOLUME,
                          existing_ref)

    def test_get_replication_status(self):
        """Test that get_replication_status return successfully. """

        self.driver.do_setup(None)

        # assume the replicated volume is inactive
        replicated_volume = copy.deepcopy(REPLICATED_VOLUME)
        replicated_volume['replication_status'] = 'inactive'
        model_update = self.driver.get_replication_status(
            CONTEXT,
            replicated_volume
        )
        self.assertEqual('active', model_update['replication_status'])

    def test_get_replication_status_fail_on_exception(self):
        """Test that get_replication_status fails on exception"""

        self.driver.do_setup(None)

        replicated_volume = copy.deepcopy(REPLICATED_VOLUME)
        # on purpose - set invalid value to replication_status
        # expect an exception.
        replicated_volume['replication_status'] = 'invalid_status_val'
        self.assertRaises(
            exception.CinderException,
            self.driver.get_replication_status,
            CONTEXT,
            replicated_volume
        )

    def test_retype(self):
        """Test that retype returns successfully."""

        self.driver.do_setup(None)

        # prepare parameters
        ctxt = context.get_admin_context()

        host = {
            'host': 'foo',
            'capabilities': {
                'location_info': 'ibm_storage_fake_1',
                'extent_size': '1024'
            }
        }

        key_specs_old = {'easytier': False, 'warning': 2, 'autoexpand': True}
        key_specs_new = {'easytier': True, 'warning': 5, 'autoexpand': False}
        old_type_ref = volume_types.create(ctxt, 'old', key_specs_old)
        new_type_ref = volume_types.create(ctxt, 'new', key_specs_new)

        diff, equal = volume_types.volume_types_diff(
            ctxt,
            old_type_ref['id'],
            new_type_ref['id'],
        )

        volume = copy.deepcopy(VOLUME)
        old_type = volume_types.get_volume_type(ctxt, old_type_ref['id'])
        volume['volume_type'] = old_type
        volume['host'] = host
        new_type = volume_types.get_volume_type(ctxt, new_type_ref['id'])

        self.driver.create_volume(volume)
        ret = self.driver.retype(ctxt, volume, new_type, diff, host)
        self.assertTrue(bool(ret))
        self.assertEqual('1', volume['easytier'])

    def test_retype_fail_on_exception(self):
        """Test that retype fails on exception."""

        self.driver.do_setup(None)

        # prepare parameters
        ctxt = context.get_admin_context()

        host = {
            'host': 'foo',
            'capabilities': {
                'location_info': 'ibm_storage_fake_1',
                'extent_size': '1024'
            }
        }

        key_specs_old = {'easytier': False, 'warning': 2, 'autoexpand': True}
        old_type_ref = volume_types.create(ctxt, 'old', key_specs_old)
        new_type_ref = volume_types.create(ctxt, 'new')

        diff, equal = volume_types.volume_types_diff(
            ctxt,
            old_type_ref['id'],
            new_type_ref['id'],
        )

        volume = copy.deepcopy(VOLUME)
        old_type = volume_types.get_volume_type(ctxt, old_type_ref['id'])
        volume['volume_type'] = old_type
        volume['host'] = host
        new_type = volume_types.get_volume_type(ctxt, new_type_ref['id'])

        self.driver.create_volume(volume)
        self.assertRaises(
            KeyError,
            self.driver.retype,
            ctxt, volume, new_type, diff, host
        )

    def test_create_group(self):
        """Test that create_group return successfully."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        model_update = self.driver.create_group(ctxt, GROUP)

        self.assertEqual(fields.GroupStatus.AVAILABLE,
                         model_update['status'],
                         "Group created failed")

    def test_create_group_fail_on_group_not_empty(self):
        """Test create_group with empty group."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create volumes
        # And add the volumes into the group before creating group
        self.driver.create_volume(GROUP_VOLUME)

        self.assertRaises(exception.CinderException,
                          self.driver.create_group,
                          ctxt, GROUP)

    def test_delete_group(self):
        """Test that delete_group return successfully."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        self.driver.create_group(ctxt, GROUP)

        # Create volumes and add them to group
        self.driver.create_volume(GROUP_VOLUME)

        # Delete group
        model_update, volumes = self.driver.delete_group(ctxt, GROUP,
                                                         [GROUP_VOLUME])

        # Verify the result
        self.assertEqual(fields.GroupStatus.DELETED,
                         model_update['status'],
                         'Group deleted failed')
        for volume in volumes:
            self.assertEqual('deleted',
                             volume['status'],
                             'Group deleted failed')

    def test_delete_group_fail_on_volume_not_delete(self):
        """Test delete_group with volume delete failure."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        self.driver.create_group(ctxt, GROUP)

        # Set the volume not to be deleted
        volume = copy.deepcopy(GROUP_VOLUME)
        volume['name'] = CANNOT_DELETE

        # Create volumes and add them to group
        self.driver.create_volume(volume)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_group,
                          ctxt, GROUP, [volume])

    def test_create_group_snapshot(self):
        """Test that create_group_snapshot return successfully."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        self.driver.create_group(ctxt, GROUP)

        # Create volumes and add them to group
        self.driver.create_volume(VOLUME)

        # Create group snapshot
        model_update, snapshots = self.driver.create_group_snapshot(
            ctxt, GROUP_SNAPSHOT, [VOLUME])

        # Verify the result
        self.assertEqual(fields.GroupSnapshotStatus.AVAILABLE,
                         model_update['status'],
                         'Group Snapshot created failed')
        for snap in snapshots:
            self.assertEqual('available',
                             snap['status'])

        # Clean the environment
        self.driver.delete_group_snapshot(ctxt, GROUP_SNAPSHOT, [VOLUME])
        self.driver.delete_group(ctxt, GROUP, [VOLUME])

    def test_create_group_snapshot_fail_on_no_pool_space_left(self):
        """Test create_group_snapshot when no pool space left."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        self.driver.create_group(ctxt, GROUP)

        # Set the volume size
        volume = copy.deepcopy(GROUP_VOLUME)
        volume['size'] = POOL_SIZE / 2 + 1

        # Create volumes and add them to group
        self.driver.create_volume(volume)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_group_snapshot,
                          ctxt, GROUP_SNAPSHOT, [volume])

        # Clean the environment
        self.driver.volumes = None
        self.driver.delete_group(ctxt, GROUP, [volume])

    def test_delete_group_snapshot(self):
        """Test that delete_group_snapshot return successfully."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        self.driver.create_group(ctxt, GROUP)

        # Create volumes and add them to group
        self.driver.create_volume(GROUP_VOLUME)

        # Create group snapshot
        self.driver.create_group_snapshot(ctxt, GROUP_SNAPSHOT, [GROUP_VOLUME])

        # Delete group snapshot
        model_update, snapshots = self.driver.delete_group_snapshot(
            ctxt, GROUP_SNAPSHOT, [GROUP_VOLUME])

        # Verify the result
        self.assertEqual(fields.GroupSnapshotStatus.DELETED,
                         model_update['status'],
                         'Group Snapshot deleted failed')
        for snap in snapshots:
            self.assertEqual('deleted',
                             snap['status'])

        # Clean the environment
        self.driver.delete_group(ctxt, GROUP, [GROUP_VOLUME])

    def test_delete_group_snapshot_fail_on_snapshot_not_delete(self):
        """Test delete_group_snapshot when the snapshot cannot be deleted."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group
        self.driver.create_group(ctxt, GROUP)

        # Set the snapshot not to be deleted
        volume = copy.deepcopy(GROUP_VOLUME)
        volume['name'] = CANNOT_DELETE

        # Create volumes and add them to group
        self.driver.create_volume(volume)

        # Create group snapshot
        self.driver.create_group_snapshot(ctxt, GROUP_SNAPSHOT, [volume])

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.delete_group_snapshot,
                          ctxt, GROUP_SNAPSHOT, [volume])

    def test_update_group_without_volumes(self):
        """Test update_group when there are no volumes specified."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Update group
        model_update, added, removed = self.driver.update_group(
            ctxt, GROUP, [], [])

        self.assertEqual(fields.GroupStatus.AVAILABLE,
                         model_update['status'],
                         "Group update failed")
        self.assertIsNone(added,
                          "added volumes list is not empty")
        self.assertIsNone(removed,
                          "removed volumes list is not empty")

    def test_update_group_with_volumes(self):
        """Test update_group when there are volumes specified."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Update group
        model_update, added, removed = self.driver.update_group(
            ctxt, GROUP, [VOLUME], [VOLUME2])

        self.assertEqual(fields.GroupStatus.AVAILABLE,
                         model_update['status'],
                         "Group update failed")
        self.assertIsNone(added,
                          "added volumes list is not empty")
        self.assertIsNone(removed,
                          "removed volumes list is not empty")

    def test_create_group_from_src_without_volumes(self):
        """Test create_group_from_src with no volumes specified."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group from source
        model_update, volumes_model_update = (
            self.driver.create_group_from_src(
                ctxt, GROUP, [], GROUP_SNAPSHOT, []))

        # model_update can be None or return available in status
        if model_update:
            self.assertEqual(fields.GroupStatus.AVAILABLE,
                             model_update['status'],
                             "Group create from source failed")
        # volumes_model_update can be None or return available in status
        if volumes_model_update:
            self.assertFalse(volumes_model_update,
                             "volumes list is not empty")

    def test_create_group_from_src_with_volumes(self):
        """Test create_group_from_src with volumes specified."""

        self.driver.do_setup(None)

        ctxt = context.get_admin_context()

        # Create group from source
        model_update, volumes_model_update = (
            self.driver.create_group_from_src(
                ctxt, GROUP, [VOLUME], GROUP_SNAPSHOT, [SNAPSHOT]))

        # model_update can be None or return available in status
        if model_update:
            self.assertEqual(fields.GroupStatus.AVAILABLE,
                             model_update['status'],
                             "Group create from source failed")
        # volumes_model_update can be None or return available in status
        if volumes_model_update:
            self.assertEqual(fields.GroupStatus.AVAILABLE,
                             volumes_model_update['status'],
                             "volumes list status failed")

    def test_freeze_backend(self):
        """Test that freeze_backend returns successful"""

        self.driver.do_setup(None)

        # not much we can test here...
        self.assertTrue(self.driver.freeze_backend(CONTEXT))

    def test_thaw_backend(self):
        """Test that thaw_backend returns successful"""

        self.driver.do_setup(None)

        # not much we can test here...
        self.assertTrue(self.driver.thaw_backend(CONTEXT))

    def test_failover_host(self):
        """Test that failover_host returns expected values"""

        self.driver.do_setup(None)

        replicated_volume = copy.deepcopy(REPLICATED_VOLUME)
        # assume the replication_status is active
        replicated_volume['replication_status'] = 'active'

        expected_target_id = 'BLA'
        expected_volume_update_list = [
            {'volume_id': REPLICATED_VOLUME['id'],
             'updates': {'replication_status': 'failed-over'}}]

        target_id, volume_update_list, __ = self.driver.failover_host(
            CONTEXT,
            [replicated_volume],
            SECONDARY,
            []
        )

        self.assertEqual(expected_target_id, target_id)
        self.assertEqual(expected_volume_update_list, volume_update_list)

    def test_failover_host_bad_state(self):
        """Test that failover_host returns with error"""

        self.driver.do_setup(None)

        replicated_volume = copy.deepcopy(REPLICATED_VOLUME)
        # assume the replication_status is active
        replicated_volume['replication_status'] = 'invalid_status_val'

        expected_target_id = 'BLA'
        expected_volume_update_list = [
            {'volume_id': REPLICATED_VOLUME['id'],
             'updates': {'replication_status': 'error'}}]

        target_id, volume_update_list, __ = self.driver.failover_host(
            CONTEXT,
            [replicated_volume],
            SECONDARY,
            []
        )

        self.assertEqual(expected_target_id, target_id)
        self.assertEqual(expected_volume_update_list, volume_update_list)

    def test_enable_replication(self):
        self.driver.do_setup(None)
        model_update, volumes_model_update = self.driver.enable_replication(
            CONTEXT, GROUP, [REPLICATED_VOLUME])
        self.assertEqual(fields.ReplicationStatus.ENABLED,
                         model_update['replication_status'])
        for vol in volumes_model_update:
            self.assertEqual(fields.ReplicationStatus.ENABLED,
                             vol['replication_status'])

    def test_disable_replication(self):
        self.driver.do_setup(None)
        model_update, volumes_model_update = self.driver.disable_replication(
            CONTEXT, GROUP, [REPLICATED_VOLUME_DISABLED])
        self.assertEqual(fields.ReplicationStatus.DISABLED,
                         model_update['replication_status'])
        for vol in volumes_model_update:
            self.assertEqual(fields.ReplicationStatus.DISABLED,
                             volumes_model_update[0]['replication_status'])

    def test_failover_replication(self):
        self.driver.do_setup(None)
        model_update, volumes_model_update = self.driver.failover_replication(
            CONTEXT, GROUP, [VOLUME], SECONDARY)
        self.assertEqual(fields.ReplicationStatus.FAILED_OVER,
                         model_update['replication_status'])

    def test_get_replication_error_status(self):
        self.driver.do_setup(None)
        group_model_updates, volume_model_updates = (
            self.driver.get_replication_error_status(CONTEXT, [GROUP]))
        self.assertEqual(fields.ReplicationStatus.ERROR,
                         group_model_updates[0]['replication_status'])
        self.assertEqual(fields.ReplicationStatus.ERROR,
                         volume_model_updates[0]['replication_status'])



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt

from cinder.api import microversions as mv
from cinder.api.v3 import snapshot_metadata
from cinder import exception
from cinder.policies import snapshot_metadata as policy
from cinder.policies import snapshots as snapshots_policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils


@ddt.ddt
class SnapshotMetadataPolicyTest(base.BasePolicyTest):
    authorized_readers = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_readers = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_members = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # DB validations will throw SnapshotNotFound for some contexts
    unauthorized_exceptions = [
        exception.SnapshotNotFound,
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = snapshot_metadata.Controller()
        self.api_path = '/v3/%s/snapshots' % (self.project_id)
        self.api_version = mv.BASE_VERSION
        self.vol_type = test_utils.create_volume_type(
            self.project_admin_context,
            name='fake_vol_type', testcase_instance=self)
        # Relax the snapshots GET_POLICY in order to get past that check.
        self.policy.set_rules({snapshots_policy.GET_POLICY: ""},
                              overwrite=False)

    def _create_volume(self, **kwargs):
        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=self.vol_type.id,
                                          testcase_instance=self, **kwargs)
        return volume

    def _create_snapshot(self, **kwargs):
        volume = self._create_volume(**kwargs)
        snapshot = test_utils.create_snapshot(self.project_member_context,
                                              volume_id=volume.id,
                                              testcase_instance=self, **kwargs)
        return snapshot

    @ddt.data(*base.all_users)
    def test_get_policy(self, user_id):
        metadata = {'inside': 'out'}
        snapshot = self._create_snapshot(metadata=metadata)
        rule_name = policy.GET_POLICY
        url = '%s/%s/metadata' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        response = self.common_policy_check(
            user_id, self.authorized_readers,
            self.unauthorized_readers,
            self.unauthorized_exceptions,
            rule_name, self.controller.show,
            req, snapshot_id=snapshot.id, id='inside')

        if user_id in self.authorized_readers:
            self.assertDictEqual(metadata, response['meta'])

    @ddt.data(*base.all_users)
    def test_update_policy(self, user_id):
        snapshot = self._create_snapshot()
        rule_name = policy.UPDATE_POLICY
        url = '%s/%s/metadata' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        metadata = {
            'inside': 'out',
            'outside': 'in'
        }
        body = {
            "metadata": {**metadata}
        }

        response = self.common_policy_check(
            user_id, self.authorized_members,
            self.unauthorized_members,
            self.unauthorized_exceptions,
            rule_name, self.controller.update_all,
            req, snapshot_id=snapshot.id, body=body)

        if user_id in self.authorized_members:
            self.assertDictEqual(metadata, response['metadata'])

    @ddt.data(*base.all_users)
    def test_delete_policy(self, user_id):
        metadata = {'inside': 'out'}
        snapshot = self._create_snapshot(metadata=metadata)
        rule_name = policy.DELETE_POLICY
        url = '%s/%s/metadata/inside' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'

        # Relax the GET_POLICY in order to get past that check.
        self.policy.set_rules({policy.GET_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.delete, req,
                                 snapshot_id=snapshot.id, id='inside')


class SnapshotMetadataPolicySecureRbacTest(SnapshotMetadataPolicyTest):
    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
    ]

    unauthorized_readers = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]

    unauthorized_members = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



#    Copyright 2016 EMC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base


@base.CinderObjectRegistry.register
class GroupSnapshot(base.CinderPersistentObject, base.CinderObject,
                    base.CinderObjectDictCompat, base.ClusteredObject):
    VERSION = '1.0'

    OPTIONAL_FIELDS = ['group', 'snapshots']

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        'id': fields.UUIDField(),
        'group_id': fields.UUIDField(nullable=False),
        'project_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'name': fields.StringField(nullable=True),
        'description': fields.StringField(nullable=True),
        'status': fields.StringField(nullable=True),
        'group_type_id': fields.UUIDField(nullable=True),
        'group': fields.ObjectField('Group', nullable=True),
        'snapshots': fields.ObjectField('SnapshotList', nullable=True),
    }

    @property
    def host(self):
        return self.group.host

    @property
    def cluster_name(self):
        return self.group.cluster_name

    @classmethod
    def _from_db_object(cls, context, group_snapshot, db_group_snapshots,
                        expected_attrs=None):
        expected_attrs = expected_attrs or []
        for name, field in group_snapshot.fields.items():
            if name in cls.OPTIONAL_FIELDS:
                continue
            value = db_group_snapshots.get(name)
            setattr(group_snapshot, name, value)

        if 'group' in expected_attrs:
            group = objects.Group(context)
            group._from_db_object(context, group,
                                  db_group_snapshots['group'])
            group_snapshot.group = group

        if 'snapshots' in expected_attrs:
            snapshots = base.obj_make_list(
                context, objects.SnapshotsList(context),
                objects.Snapshots,
                db_group_snapshots['snapshots'])
            group_snapshot.snapshots = snapshots

        group_snapshot._context = context
        group_snapshot.obj_reset_changes()
        return group_snapshot

    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already_created'))
        updates = self.cinder_obj_get_changes()

        if 'group' in updates:
            raise exception.ObjectActionError(
                action='create', reason=_('group assigned'))

        db_group_snapshots = db.group_snapshot_create(self._context, updates)
        self._from_db_object(self._context, self, db_group_snapshots)

    def obj_load_attr(self, attrname):
        if attrname not in self.OPTIONAL_FIELDS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        if attrname == 'group':
            self.group = objects.Group.get_by_id(
                self._context, self.group_id)

        if attrname == 'snapshots':
            self.snapshots = objects.SnapshotList.get_all_for_group_snapshot(
                self._context, self.id)

        self.obj_reset_changes(fields=[attrname])

    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            if 'group' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('group changed'))
            if 'snapshots' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('snapshots changed'))
            db.group_snapshot_update(self._context, self.id, updates)
            self.obj_reset_changes()

    def destroy(self):
        with self.obj_as_admin():
            updated_values = db.group_snapshot_destroy(self._context, self.id)
        self.update(updated_values)
        self.obj_reset_changes(updated_values.keys())


@base.CinderObjectRegistry.register
class GroupSnapshotList(base.ObjectListBase, base.CinderObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('GroupSnapshot')
    }

    @classmethod
    def get_all(cls, context, filters=None, marker=None, limit=None,
                offset=None, sort_keys=None, sort_dirs=None):
        group_snapshots = db.group_snapshot_get_all(context,
                                                    filters=filters,
                                                    marker=marker,
                                                    limit=limit,
                                                    offset=offset,
                                                    sort_keys=sort_keys,
                                                    sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context), objects.GroupSnapshot,
                                  group_snapshots)

    @classmethod
    def get_all_by_project(cls, context, project_id, filters=None, marker=None,
                           limit=None, offset=None, sort_keys=None,
                           sort_dirs=None):
        group_snapshots = db.group_snapshot_get_all_by_project(
            context, project_id, filters=filters, marker=marker,
            limit=limit, offset=offset, sort_keys=sort_keys,
            sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context), objects.GroupSnapshot,
                                  group_snapshots)

    @classmethod
    def get_all_by_group(cls, context, group_id, filters=None, marker=None,
                         limit=None, offset=None, sort_keys=None,
                         sort_dirs=None):
        group_snapshots = db.group_snapshot_get_all_by_group(
            context, group_id, filters=filters, marker=marker, limit=limit,
            offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context), objects.GroupSnapshot,
                                  group_snapshots)



# Copyright (c) 2012 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Configuration support for all drivers.

This module allows support for setting configurations either from default
or from a particular FLAGS group, to be able to set multiple configurations
for a given set of values.

For instance, two lvm configurations can be set by naming them in groups as

 [lvm1]
 volume_group=lvm-group-1
 ...

 [lvm2]
 volume_group=lvm-group-2
 ...

And the configuration group name will be passed in so that all calls to
configuration.volume_group within that instance will be mapped to the proper
named group.

This class also ensures the implementation's configuration is grafted into the
option group. This is due to the way cfg works. All cfg options must be defined
and registered in the group in which they are used.
"""


from oslo_config import cfg


CONF = cfg.CONF
SHARED_CONF_GROUP = 'backend_defaults'


class DefaultGroupConfiguration(object):
    """Get config options from only DEFAULT."""

    def __init__(self):
        # set the local conf so that __call__'s know what to use
        self.local_conf = CONF

    def _ensure_config_values(self, volume_opts):
        CONF.register_opts(volume_opts, group=None)

    def append_config_values(self, volume_opts):
        self._ensure_config_values(volume_opts)

    def safe_get(self, value):
        try:
            return self.__getattr__(value)
        except cfg.NoSuchOptError:
            return None

    def __getattr__(self, value):
        # Don't use self.local_conf to avoid reentrant call to __getattr__()
        local_conf = object.__getattribute__(self, 'local_conf')
        return getattr(local_conf, value)


class BackendGroupConfiguration(object):

    def __init__(self, volume_opts, config_group=None):
        """Initialize configuration.

        This takes care of grafting the implementation's config
        values into the config group and shared defaults. We will try to
        pull values from the specified 'config_group', but fall back to
        defaults from the SHARED_CONF_GROUP.
        """
        self.config_group = config_group

        # set the local conf so that __call__'s know what to use
        self._ensure_config_values(volume_opts)
        self.backend_conf = CONF._get(self.config_group)
        self.shared_backend_conf = CONF._get(SHARED_CONF_GROUP)

    def _safe_register(self, opt, group):
        try:
            CONF.register_opt(opt, group=group)
        except cfg.DuplicateOptError:
            pass  # If it's already registered ignore it

    def _ensure_config_values(self, volume_opts):
        """Register the options in the shared group.

        When we go to get a config option we will try the backend specific
        group first and fall back to the shared group. We override the default
        from all the config options for the backend group so we can know if it
        was set or not.
        """
        for opt in volume_opts:
            self._safe_register(opt, SHARED_CONF_GROUP)
            # Assuming they aren't the same groups, graft on the options into
            # the backend group and override its default value.
            if self.config_group != SHARED_CONF_GROUP:
                self._safe_register(opt, self.config_group)
                CONF.set_default(opt.name, None, group=self.config_group)

    def append_config_values(self, volume_opts):
        self._ensure_config_values(volume_opts)

    def set_default(self, opt_name, default):
        CONF.set_default(opt_name, default, group=SHARED_CONF_GROUP)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def safe_get(self, value):
        try:
            return self.__getattr__(value)
        except cfg.NoSuchOptError:
            return None

    def __getattr__(self, opt_name):
        # Don't use self.X to avoid reentrant call to __getattr__()
        backend_conf = object.__getattribute__(self, 'backend_conf')
        opt_value = getattr(backend_conf, opt_name)
        if opt_value is None:
            shared_conf = object.__getattribute__(self, 'shared_backend_conf')
            opt_value = getattr(shared_conf, opt_name)
        return opt_value


class Configuration(object):

    def __init__(self, volume_opts, config_group=None):
        """Initialize configuration.

        This shim will allow for compatibility with the DEFAULT
        style of backend configuration which is used by some of the users
        of this configuration helper, or by the volume drivers that have
        all been forced over to the config_group style.
        """
        self.config_group = config_group
        if config_group:
            self.conf = BackendGroupConfiguration(volume_opts, config_group)
        else:
            self.conf = DefaultGroupConfiguration()

    def append_config_values(self, volume_opts):
        self.conf.append_config_values(volume_opts)

    def safe_get(self, value):
        return self.conf.safe_get(value)

    def __getattr__(self, value):
        # Don't use self.conf to avoid reentrant call to __getattr__()
        conf = object.__getattribute__(self, 'conf')
        return getattr(conf, value)



#    Copyright 2014 Objectif Libre
#    Copyright 2015 Dot Hill Systems Corp.
#    Copyright 2016-2019 Seagate Technology or one of its affiliates
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from cinder.common import constants
import cinder.volume.driver
import cinder.volume.drivers.san.san as san
import cinder.volume.drivers.stx.common as common
from cinder.zonemanager import utils as fczm_utils


class STXFCDriver(cinder.volume.driver.FibreChannelDriver):
    """OpenStack Fibre Channel cinder drivers for Seagate arrays.

    .. code:: text

      Version history:
          0.1    - Base version developed for HPMSA FC drivers:
                      "https://github.com/openstack/cinder/tree/stable/juno/
                       cinder/volume/drivers/san/hp"
          1.0    - Version developed for DotHill arrays with the following
                   modifications:
                       - added support for v3 API(virtual pool feature)
                       - added support for retype volume
                       - added support for manage/unmanage volume
                       - added initiator target mapping in FC zoning
                       - added https support
          1.6    - Add management path redundancy and reduce load placed
                   on management controller.
          1.7    - Modified so it can't be invoked except as a superclass
          2.0    - Reworked to create a new Seagate (STX) array driver.
    """

    VERSION = "2.0"

    CI_WIKI_NAME = 'Seagate_CI'

    def __init__(self, *args, **kwargs):
        super(STXFCDriver, self).__init__(*args, **kwargs)
        self.common = None
        self.configuration.append_config_values(san.san_opts)
        self.lookup_service = fczm_utils.create_lookup_service()
        if type(self) != STXFCDriver:
            return
        self.configuration.append_config_values(common.common_opts)

    def _init_common(self):
        return common.STXCommon(self.configuration)

    def _check_flags(self):
        required_flags = ['san_ip', 'san_login', 'san_password']
        self.common.check_flags(self.configuration, required_flags)

    def do_setup(self, context):
        self.common = self._init_common()
        self._check_flags()
        self.common.do_setup(context)

    def check_for_setup_error(self):
        self._check_flags()

    def create_volume(self, volume):
        self.common.create_volume(volume)

    def create_volume_from_snapshot(self, volume, src_vref):
        self.common.create_volume_from_snapshot(volume, src_vref)

    def create_cloned_volume(self, volume, src_vref):
        self.common.create_cloned_volume(volume, src_vref)

    def delete_volume(self, volume):
        self.common.delete_volume(volume)

    def initialize_connection(self, volume, connector):
        self.common.client_login()
        try:
            data = {}
            data['target_lun'] = self.common.map_volume(volume,
                                                        connector,
                                                        'wwpns')

            ports, init_targ_map = self.get_init_targ_map(connector)
            data['target_discovered'] = True
            data['target_wwn'] = ports
            data['initiator_target_map'] = init_targ_map
            info = {'driver_volume_type': 'fibre_channel',
                    'data': data}
            fczm_utils.add_fc_zone(info)
            return info
        finally:
            self.common.client_logout()

    def terminate_connection(self, volume, connector, **kwargs):
        info = {'driver_volume_type': 'fibre_channel', 'data': {}}
        try:
            if not self.common.client.list_luns_for_host(
                    connector['wwpns'][0]):
                ports, init_targ_map = self.get_init_targ_map(connector)
                info['data'] = {'target_wwn': ports,
                                'initiator_target_map': init_targ_map}
            # multiattach volumes cannot be unmapped here, but will
            # be implicity unmapped when the volume is deleted.
            if not volume.get('multiattach'):
                self.common.unmap_volume(volume, connector, 'wwpns')
                fczm_utils.remove_fc_zone(info)
        finally:
            return info

    def get_init_targ_map(self, connector):
        init_targ_map = {}
        target_wwns = []
        ports = self.common.get_active_fc_target_ports()
        if self.lookup_service is not None:
            dev_map = self.lookup_service.get_device_mapping_from_network(
                connector['wwpns'],
                ports)
            for fabric_name in dev_map:
                fabric = dev_map[fabric_name]
                target_wwns += fabric['target_port_wwn_list']
                for initiator in fabric['initiator_port_wwn_list']:
                    if initiator not in init_targ_map:
                        init_targ_map[initiator] = []
                    init_targ_map[initiator] += fabric['target_port_wwn_list']
                    init_targ_map[initiator] = list(set(
                                                    init_targ_map[initiator]))
            target_wwns = list(set(target_wwns))
        else:
            initiator_wwns = connector['wwpns']
            target_wwns = ports
            for initiator in initiator_wwns:
                init_targ_map[initiator] = target_wwns

        return target_wwns, init_targ_map

    def get_volume_stats(self, refresh=False):
        stats = self.common.get_volume_stats(refresh)
        stats['storage_protocol'] = constants.FC
        stats['driver_version'] = self.VERSION
        backend_name = self.configuration.safe_get('volume_backend_name')
        stats['volume_backend_name'] = (backend_name or
                                        self.__class__.__name__)
        return stats

    def create_export(self, context, volume, connector=None):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    def create_snapshot(self, snapshot):
        self.common.create_snapshot(snapshot)

    def delete_snapshot(self, snapshot):
        self.common.delete_snapshot(snapshot)

    def extend_volume(self, volume, new_size):
        self.common.extend_volume(volume, new_size)

    def retype(self, context, volume, new_type, diff, host):
        return self.common.retype(volume, new_type, diff, host)

    def manage_existing(self, volume, existing_ref):
        self.common.manage_existing(volume, existing_ref)

    def manage_existing_get_size(self, volume, existing_ref):
        return self.common.manage_existing_get_size(volume, existing_ref)

    def manage_existing_snapshot(self, snapshot, existing_ref):
        return self.common.manage_existing_snapshot(snapshot, existing_ref)

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        return self.common.manage_existing_snapshot_get_size(snapshot,
                                                             existing_ref)

    def unmanage(self, volume):
        pass

    def unmanage_snapshot(self, snapshot):
        pass

    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        return self.common.get_manageable_volumes(cinder_volumes,
                                                  marker, limit, offset,
                                                  sort_keys, sort_dirs)

    def get_manageable_snapshots(self, cinder_snapshots, marker, limit,
                                 offset, sort_keys, sort_dirs):
        return self.common.get_manageable_snapshots(cinder_snapshots,
                                                    marker, limit, offset,
                                                    sort_keys, sort_dirs)

    @staticmethod
    def get_driver_options():
        return common.STXCommon.get_driver_options()



# Copyright (C) 2021 NEC corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""REST interface for NEC Driver."""

from oslo_config import cfg

from cinder.volume import configuration
from cinder.volume.drivers.hitachi import hbsd_rest
from cinder.volume.drivers.hitachi import hbsd_rest_api
from cinder.volume.drivers.hitachi import hbsd_rest_fc
from cinder.volume.drivers.hitachi import hbsd_rest_iscsi

COMMON_VOLUME_OPTS = [
    cfg.StrOpt(
        'nec_v_storage_id',
        default=None,
        help='Product number of the storage system.'),
    cfg.StrOpt(
        'nec_v_pool',
        default=None,
        help='Pool number or pool name of the DP pool.'),
    cfg.StrOpt(
        'nec_v_snap_pool',
        default=None,
        help='Pool number or pool name of the snapshot pool.'),
    cfg.StrOpt(
        'nec_v_ldev_range',
        default=None,
        help='Range of the LDEV numbers in the format of \'xxxx-yyyy\' that '
             'can be used by the driver. Values can be in decimal format '
             '(e.g. 1000) or in colon-separated hexadecimal format '
             '(e.g. 00:03:E8).'),
    cfg.ListOpt(
        'nec_v_target_ports',
        default=[],
        help='IDs of the storage ports used to attach volumes to the '
             'controller node. To specify multiple ports, connect them by '
             'commas (e.g. CL1-A,CL2-A).'),
    cfg.ListOpt(
        'nec_v_compute_target_ports',
        default=[],
        help='IDs of the storage ports used to attach volumes to compute '
             'nodes. To specify multiple ports, connect them by commas '
             '(e.g. CL1-A,CL2-A).'),
    cfg.BoolOpt(
        'nec_v_group_create',
        default=False,
        help='If True, the driver will create host groups or iSCSI targets on '
             'storage ports as needed.'),
    cfg.BoolOpt(
        'nec_v_group_delete',
        default=False,
        help='If True, the driver will delete host groups or iSCSI targets on '
             'storage ports as needed.'),
    cfg.IntOpt(
        'nec_v_copy_speed',
        default=3,
        min=1, max=15,
        help='Copy speed of storage system. 1 or 2 indicates '
             'low speed, 3 indicates middle speed, and a value between 4 and '
             '15 indicates high speed.'),
    cfg.IntOpt(
        'nec_v_copy_check_interval',
        default=3,
        min=1, max=600,
        help='Interval in seconds to check copying status during a volume '
             'copy.'),
    cfg.IntOpt(
        'nec_v_async_copy_check_interval',
        default=10,
        min=1, max=600,
        help='Interval in seconds to check asynchronous copying status during '
             'a copy pair deletion or data restoration.'),
]

REST_VOLUME_OPTS = [
    cfg.BoolOpt(
        'nec_v_rest_disable_io_wait',
        default=True,
        help='It may take some time to detach volume after I/O. '
             'This option will allow detaching volume to complete '
             'immediately.'),
    cfg.BoolOpt(
        'nec_v_rest_tcp_keepalive',
        default=True,
        help='Enables or disables use of REST API tcp keepalive'),
    cfg.BoolOpt(
        'nec_v_discard_zero_page',
        default=True,
        help='Enable or disable zero page reclamation in a DP-VOL.'),
    cfg.IntOpt(
        'nec_v_lun_timeout',
        default=hbsd_rest._LUN_TIMEOUT,
        help='Maximum wait time in seconds for adding a LUN to complete.'),
    cfg.IntOpt(
        'nec_v_lun_retry_interval',
        default=hbsd_rest._LUN_RETRY_INTERVAL,
        help='Retry interval in seconds for REST API adding a LUN.'),
    cfg.IntOpt(
        'nec_v_restore_timeout',
        default=hbsd_rest._RESTORE_TIMEOUT,
        help='Maximum wait time in seconds for the restore operation to '
             'complete.'),
    cfg.IntOpt(
        'nec_v_state_transition_timeout',
        default=hbsd_rest._STATE_TRANSITION_TIMEOUT,
        help='Maximum wait time in seconds for a volume transition to '
             'complete.'),
    cfg.IntOpt(
        'nec_v_lock_timeout',
        default=hbsd_rest_api._LOCK_TIMEOUT,
        help='Maximum wait time in seconds for storage to be unlocked.'),
    cfg.IntOpt(
        'nec_v_rest_timeout',
        default=hbsd_rest_api._REST_TIMEOUT,
        help='Maximum wait time in seconds for REST API execution to '
             'complete.'),
    cfg.IntOpt(
        'nec_v_extend_timeout',
        default=hbsd_rest_api._EXTEND_TIMEOUT,
        help='Maximum wait time in seconds for a volume extention to '
             'complete.'),
    cfg.IntOpt(
        'nec_v_exec_retry_interval',
        default=hbsd_rest_api._EXEC_RETRY_INTERVAL,
        help='Retry interval in seconds for REST API execution.'),
    cfg.IntOpt(
        'nec_v_rest_connect_timeout',
        default=hbsd_rest_api._DEFAULT_CONNECT_TIMEOUT,
        help='Maximum wait time in seconds for REST API connection to '
             'complete.'),
    cfg.IntOpt(
        'nec_v_rest_job_api_response_timeout',
        default=hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT,
        help='Maximum wait time in seconds for a response from REST API.'),
    cfg.IntOpt(
        'nec_v_rest_get_api_response_timeout',
        default=hbsd_rest_api._GET_API_RESPONSE_TIMEOUT,
        help='Maximum wait time in seconds for a response against GET method '
             'of REST API.'),
    cfg.IntOpt(
        'nec_v_rest_server_busy_timeout',
        default=hbsd_rest_api._REST_SERVER_BUSY_TIMEOUT,
        help='Maximum wait time in seconds when REST API returns busy.'),
    cfg.IntOpt(
        'nec_v_rest_keep_session_loop_interval',
        default=hbsd_rest_api._KEEP_SESSION_LOOP_INTERVAL,
        help='Loop interval in seconds for keeping REST API session.'),
    cfg.IntOpt(
        'nec_v_rest_another_ldev_mapped_retry_timeout',
        default=hbsd_rest_api._ANOTHER_LDEV_MAPPED_RETRY_TIMEOUT,
        help='Retry time in seconds when new LUN allocation request fails.'),
    cfg.IntOpt(
        'nec_v_rest_tcp_keepidle',
        default=hbsd_rest_api._TCP_KEEPIDLE,
        help='Wait time in seconds for sending a first TCP keepalive packet.'),
    cfg.IntOpt(
        'nec_v_rest_tcp_keepintvl',
        default=hbsd_rest_api._TCP_KEEPINTVL,
        help='Interval of transmissions in seconds for TCP keepalive packet.'),
    cfg.IntOpt(
        'nec_v_rest_tcp_keepcnt',
        default=hbsd_rest_api._TCP_KEEPCNT,
        help='Maximum number of transmissions for TCP keepalive packet.'),
    cfg.ListOpt(
        'nec_v_host_mode_options',
        default=[],
        help='Host mode option for host group or iSCSI target'),
]

FC_VOLUME_OPTS = [
    cfg.BoolOpt(
        'nec_v_zoning_request',
        default=False,
        help='If True, the driver will configure FC zoning between the server '
             'and the storage system provided that FC zoning manager is '
             'enabled.'),
]

CONF = cfg.CONF
CONF.register_opts(COMMON_VOLUME_OPTS, group=configuration.SHARED_CONF_GROUP)
CONF.register_opts(REST_VOLUME_OPTS, group=configuration.SHARED_CONF_GROUP)
CONF.register_opts(FC_VOLUME_OPTS, group=configuration.SHARED_CONF_GROUP)


def update_conf(conf):
    # COMMON_VOLUME_OPTS
    conf.hitachi_storage_id = conf.nec_v_storage_id
    conf.hitachi_pool = conf.nec_v_pool
    conf.hitachi_snap_pool = conf.nec_v_snap_pool
    conf.hitachi_ldev_range = conf.nec_v_ldev_range
    conf.hitachi_target_ports = conf.nec_v_target_ports
    conf.hitachi_compute_target_ports = (
        conf.nec_v_compute_target_ports)
    conf.hitachi_group_create = conf.nec_v_group_create
    conf.hitachi_group_delete = conf.nec_v_group_delete
    conf.hitachi_copy_speed = conf.nec_v_copy_speed
    conf.hitachi_copy_check_interval = (
        conf.nec_v_copy_check_interval)
    conf.hitachi_async_copy_check_interval = (
        conf.nec_v_async_copy_check_interval)

    # REST_VOLUME_OPTS
    conf.hitachi_rest_disable_io_wait = (
        conf.nec_v_rest_disable_io_wait)
    conf.hitachi_rest_tcp_keepalive = (
        conf.nec_v_rest_tcp_keepalive)
    conf.hitachi_discard_zero_page = (
        conf.nec_v_discard_zero_page)
    conf.hitachi_lun_timeout = conf.nec_v_lun_timeout
    conf.hitachi_lun_retry_interval = (
        conf.nec_v_lun_retry_interval)
    conf.hitachi_restore_timeout = conf.nec_v_restore_timeout
    conf.hitachi_state_transition_timeout = (
        conf.nec_v_state_transition_timeout)
    conf.hitachi_lock_timeout = conf.nec_v_lock_timeout
    conf.hitachi_rest_timeout = conf.nec_v_rest_timeout
    conf.hitachi_extend_timeout = conf.nec_v_extend_timeout
    conf.hitachi_exec_retry_interval = (
        conf.nec_v_exec_retry_interval)
    conf.hitachi_rest_connect_timeout = (
        conf.nec_v_rest_connect_timeout)
    conf.hitachi_rest_job_api_response_timeout = (
        conf.nec_v_rest_job_api_response_timeout)
    conf.hitachi_rest_get_api_response_timeout = (
        conf.nec_v_rest_get_api_response_timeout)
    conf.hitachi_rest_server_busy_timeout = (
        conf.nec_v_rest_server_busy_timeout)
    conf.hitachi_rest_keep_session_loop_interval = (
        conf.nec_v_rest_keep_session_loop_interval)
    conf.hitachi_rest_another_ldev_mapped_retry_timeout = (
        conf.nec_v_rest_another_ldev_mapped_retry_timeout)
    conf.hitachi_rest_tcp_keepidle = (
        conf.nec_v_rest_tcp_keepidle)
    conf.hitachi_rest_tcp_keepintvl = (
        conf.nec_v_rest_tcp_keepintvl)
    conf.hitachi_rest_tcp_keepcnt = (
        conf.nec_v_rest_tcp_keepcnt)
    conf.hitachi_host_mode_options = (
        conf.nec_v_host_mode_options)

    return conf


class VStorageRESTFC(hbsd_rest_fc.HBSDRESTFC):
    """REST interface fibre channel class."""

    def __init__(self, conf, storage_protocol, db):
        """Initialize instance variables."""
        conf.append_config_values(COMMON_VOLUME_OPTS)
        conf.append_config_values(REST_VOLUME_OPTS)
        conf.append_config_values(FC_VOLUME_OPTS)
        super(VStorageRESTFC, self).__init__(conf, storage_protocol, db)
        self.conf = update_conf(self.conf)
        # FC_VOLUME_OPTS
        self.conf.hitachi_zoning_request = self.conf.nec_v_zoning_request


class VStorageRESTISCSI(hbsd_rest_iscsi.HBSDRESTISCSI):
    """REST interface iSCSI channel class."""

    def __init__(self, conf, storage_protocol, db):
        """Initialize instance variables."""
        conf.append_config_values(COMMON_VOLUME_OPTS)
        conf.append_config_values(REST_VOLUME_OPTS)
        super(VStorageRESTISCSI, self).__init__(conf, storage_protocol, db)
        self.conf = update_conf(self.conf)
